import copy
from pathlib import Path
from unittest import TestCase

from musurgia.musurgia_exceptions import (
    SegmentedLineSegmentHasMarginsError,
    SegmentedLineLengthsCannotBeSetError,
)
from musurgia.pdf import DrawObjectRow
from musurgia.pdf.line import HorizontalSegmentedLine
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.pdf_tools import draw_ruler
from musurgia.tests.utils_for_tests import PdfTestCase

path = Path(__file__)


def change_markline_lengths(hsl):
    for index, sg in enumerate(hsl.segments):
        sg.start_mark_line.length += index * 5
    hsl.segments[-1].end_mark_line.length = hsl.segments[-1].start_mark_line.length + 5
    hsl.segments[-1].end_mark_line.show = True


class TestHorizontalSegmentedLine(TestCase):
    def setUp(self) -> None:
        self.hsl = HorizontalSegmentedLine(lengths=[10, 15, 20, 25])

    def test_segments(self):
        assert [seg.length for seg in self.hsl.segments] == [10, 15, 20, 25]
        assert self.hsl.segments[-1].end_mark_line.show is True
        assert set([seg.end_mark_line.show for seg in self.hsl.segments[:-1]]) == {
            False
        }
        assert set([seg.start_mark_line.show for seg in self.hsl.segments]) == {True}

    def test_lengths(self):
        with self.assertRaises(SegmentedLineLengthsCannotBeSetError):
            self.hsl.lengths = [10, 15, 20, 25]
        with self.assertRaises(AttributeError):
            assert self.hsl.lengths == [10, 15, 20, 25]
        assert self.hsl.get_lengths() == [10, 15, 20, 25]

    def test_get_height(self):
        self.hsl.segments[1].start_mark_line.length = 5
        assert self.hsl.get_height() == 5

    def test_set_straight_line_y_x(self):
        self.hsl.set_straight_line_relative_y(0)
        assert self.hsl.relative_y == -1.5
        with self.assertRaises(NotImplementedError):
            self.hsl.set_straight_line_relative_x(0)

    def test_align_segments(self):
        change_markline_lengths(self.hsl)
        self.hsl._align_segments()
        for seg in self.hsl.segments:
            assert seg.relative_y >= 0
            assert seg.relative_x >= 0
        assert 0 in set([seg.relative_y for seg in self.hsl.segments])

    def test_no_margin_error(self):
        self.hsl.segments[0].margins = (1, 2, 3, 4)
        with self.assertRaises(SegmentedLineSegmentHasMarginsError):
            self.hsl.draw(Pdf())

    def test_relative_y(self):
        assert isinstance(self.hsl, DrawObjectRow)
        assert self.hsl.positions == (0, 0)
        assert self.hsl.get_end_positions() == (70, 3)
        self.hsl.relative_y = 10
        assert self.hsl.positions == (0, 10)
        assert self.hsl.get_end_positions() == (70, 13)

    def test_borders(self):
        assert self.hsl.get_border_rectangle_coordinates() == (0, 0, 70, 3)

        self.hsl.set_straight_line_relative_y(0)
        assert self.hsl.get_border_rectangle_coordinates() == (0, -1.5, 70, 3)

        self.hsl.relative_y = -5
        assert self.hsl.get_border_rectangle_coordinates() == (0, -5, 70, 3)

        self.hsl.relative_y = 5
        assert self.hsl.get_border_rectangle_coordinates() == (0, 5, 70, 3)


class TestHorizontalSegmentedLineDraw(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.hsl = HorizontalSegmentedLine(lengths=[10, 15, 20, 25])

    def test_draw(self):
        with self.file_path(parent_path=path, name="draw", extension="pdf") as pdf_path:
            self.pdf.translate_page_margins()
            self.hsl.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_add_label_left(self):
        self.hsl.add_label("first left label", placement="left")
        self.hsl.add_label("second left label", placement="left")
        self.hsl.add_label("third left label", placement="left")
        self.hsl.add_label("fourth left label", placement="left")
        self.hsl.segments[0].start_mark_line.length = 10
        self.pdf.translate_page_margins()
        self.pdf.translate(40, 20)
        self.hsl.draw(self.pdf)
        with self.file_path(path, "add_label_left", "pdf") as pdf_path:
            self.pdf.write_to_path(pdf_path)

    def test_change_markline_lengths(self):
        assert set(
            [sg.start_mark_line.length for sg in self.hsl.segments]
            + [sg.end_mark_line.length for sg in self.hsl.segments]
        ) == {3}
        change_markline_lengths(self.hsl)

        assert set([sg.relative_y for sg in self.hsl.segments]) == {0}

        with self.file_path(path, "change_markline_lengths", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.translate(0, 30)
            for sg in self.hsl.segments:
                copied = copy.deepcopy(sg)
                copied.straight_line.add_text_label(
                    f"{copied.straight_line.get_positions()}",
                    font_size=5,
                    placement="below",
                )
                copied.start_mark_line.add_text_label(
                    f"{copied.start_mark_line.get_positions()}",
                    font_size=5,
                    placement="above",
                )
                copied.draw(self.pdf)
                self.pdf.translate(30, 0)
            self.pdf.reset_position()
            self.pdf.translate_page_margins()
            self.pdf.translate(0, 60)
            self.hsl._align_segments()
            self.hsl.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_margins(self):
        copy_1 = copy.deepcopy(self.hsl)
        copy_1.margins = (10, 10, 10, 10)
        dos = [self.hsl, copy_1]
        for do in dos:
            do.show_borders = True
            do.show_margins = True
        with self.file_path(
            parent_path=path, name="margins", extension="pdf"
        ) as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, mode="v")
            draw_ruler(self.pdf, mode="h")
            self.pdf.translate(20, 20)
            for do in dos:
                do.draw(self.pdf)
                self.pdf.translate(0, 20)
            self.pdf.write_to_path(pdf_path)

    def test_borders(self):
        def add_positions(do):
            do.add_text_label(f"{do.positions}", placement="left", font_size=8)

        copied_1 = copy.deepcopy(self.hsl)
        copied_2 = copy.deepcopy(self.hsl)
        copied_3 = copy.deepcopy(self.hsl)

        copied_1.set_straight_line_relative_y(0)
        copied_2.relative_y = -5
        copied_3.relative_y = 5

        assert isinstance(copied_3, DrawObjectRow)

        assert self.hsl.positions == (0, 0)
        assert [seg.positions for seg in self.hsl.segments] == [
            (0, 0),
            (0, 0),
            (0, 0),
            (0, 0),
        ]
        assert self.hsl.get_end_positions() == (70, 3)
        assert [seg.get_end_positions() for seg in self.hsl.segments] == [
            (10, 3),
            (15, 3),
            (20, 3),
            (25, 3),
        ]

        assert copied_1.positions == (0, -1.5)
        assert copied_1.get_end_positions() == (70, 1.5)
        assert [seg.positions for seg in copied_1.segments] == [
            (0, 0),
            (0, 0),
            (0, 0),
            (0, 0),
        ]
        assert [seg.get_end_positions() for seg in copied_1.segments] == [
            (10, 3),
            (15, 3),
            (20, 3),
            (25, 3),
        ]

        assert copied_2.positions == (0, -5)
        assert copied_2.get_end_positions() == (70, -2)

        assert copied_3.positions == (0, 5)
        assert copied_3.get_end_positions() == (70, 8)

        dos = [self.hsl, copied_1, copied_2, copied_3]
        for do in dos:
            do.show_borders = True

            add_positions(do)

        with self.file_path(
            parent_path=path, name="borders", extension="pdf"
        ) as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, mode="v")
            draw_ruler(self.pdf, mode="h")
            self.pdf.translate(20, 20)
            for do in dos:
                do.draw(self.pdf)
                self.pdf.translate(0, 20)
            self.pdf.write_to_path(pdf_path)



from pathlib import Path
from unittest import TestCase

from musicscore.score import Score
from musurgia.tests.utils_for_tests import XMLTestCase, create_test_fractal_timline_tree
from musurgia.trees.timelinetree import SimpleTimelineChordFactory

path = Path(__file__)


class SimpleFractalTreeChordFactory(SimpleTimelineChordFactory):
    pass


class TestSimpleFractalTreeChordFactory(TestCase):
    def setUp(self) -> None:
        self.ft = create_test_fractal_timline_tree()

    def test_node(self):
        sftchf = SimpleFractalTreeChordFactory(timline_node=self.ft)
        print(sftchf._CHORD_UPDATE_METHODS)
        assert sftchf._timeline_node == self.ft
        chord = sftchf.create_chord()
        assert (
            chord.quarter_duration.value
            == self.ft.get_duration().calculate_in_seconds()
        )


class TestFtToScore(XMLTestCase):
    def setUp(self) -> None:
        self.ft = create_test_fractal_timline_tree()
        self.score = Score()

    def test_fractal_timeline_tree_durations(self):
        expected = """└── 10
    ├── 3
    │   ├── 3/5
    │   ├── 6/5
    │   │   ├── 6/25
    │   │   ├── 12/25
    │   │   ├── 3/25
    │   │   └── 9/25
    │   ├── 3/10
    │   └── 9/10
    │       ├── 27/100
    │       ├── 9/100
    │       ├── 9/25
    │       └── 9/50
    ├── 1
    ├── 4
    │   ├── 2/5
    │   ├── 4/5
    │   ├── 6/5
    │   │   ├── 6/25
    │   │   ├── 12/25
    │   │   ├── 3/25
    │   │   └── 9/25
    │   └── 8/5
    │       ├── 4/25
    │       ├── 8/25
    │       ├── 12/25
    │       └── 16/25
    └── 2
        ├── 4/5
        │   ├── 4/25
        │   ├── 8/25
        │   ├── 2/25
        │   └── 6/25
        ├── 3/5
        │   ├── 9/50
        │   ├── 3/50
        │   ├── 6/25
        │   └── 3/25
        ├── 2/5
        └── 1/5
"""
        self.assertEqual(
            self.ft.get_tree_representation(
                key=lambda node: node.get_duration().calculate_in_seconds()
            ),
            expected,
        )
        expected = """└── QuarterDuration: 10
    ├── QuarterDuration: 3
    │   ├── QuarterDuration: 3/5
    │   ├── QuarterDuration: 6/5
    │   │   ├── QuarterDuration: 6/25
    │   │   ├── QuarterDuration: 12/25
    │   │   ├── QuarterDuration: 3/25
    │   │   └── QuarterDuration: 9/25
    │   ├── QuarterDuration: 3/10
    │   └── QuarterDuration: 9/10
    │       ├── QuarterDuration: 27/100
    │       ├── QuarterDuration: 9/100
    │       ├── QuarterDuration: 9/25
    │       └── QuarterDuration: 9/50
    ├── QuarterDuration: 1
    ├── QuarterDuration: 4
    │   ├── QuarterDuration: 2/5
    │   ├── QuarterDuration: 4/5
    │   ├── QuarterDuration: 6/5
    │   │   ├── QuarterDuration: 6/25
    │   │   ├── QuarterDuration: 12/25
    │   │   ├── QuarterDuration: 3/25
    │   │   └── QuarterDuration: 9/25
    │   └── QuarterDuration: 8/5
    │       ├── QuarterDuration: 4/25
    │       ├── QuarterDuration: 8/25
    │       ├── QuarterDuration: 12/25
    │       └── QuarterDuration: 16/25
    └── QuarterDuration: 2
        ├── QuarterDuration: 4/5
        │   ├── QuarterDuration: 4/25
        │   ├── QuarterDuration: 8/25
        │   ├── QuarterDuration: 2/25
        │   └── QuarterDuration: 6/25
        ├── QuarterDuration: 3/5
        │   ├── QuarterDuration: 9/50
        │   ├── QuarterDuration: 3/50
        │   ├── QuarterDuration: 6/25
        │   └── QuarterDuration: 3/25
        ├── QuarterDuration: 2/5
        └── QuarterDuration: 1/5
"""
        self.assertEqual(
            self.ft.get_tree_representation(
                key=lambda node: node.get_duration().get_quarter_duration()
            ),
            expected,
        )

    def test_fractal_timeline_tree_layers_to_score_simple(self):
        _show_metronome = True
        for layer_number in range(self.ft.get_number_of_layers() + 1):
            part = self.score.add_part(f"part-{layer_number + 1}")
            layer = self.ft.get_layer(level=layer_number)
            for node in layer:
                chord = SimpleFractalTreeChordFactory(
                    node, show_metronome=_show_metronome
                ).create_chord()
                part.add_chord(chord)
                _show_metronome = False
        self.score.set_possible_subdivisions([2, 3, 4, 5, 6, 7, 8])
        self.score.get_quantized = True
        self.score.finalize()
        with self.file_path(path, "simple") as xml_path:
            self.score.export_xml(xml_path)



from fractions import Fraction
from unittest import TestCase

from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.musurgia_exceptions import DAndSError


class TestArithmeticProgression(TestCase):
    def test_a1(self):
        arith = ArithmeticProgression(n=3, an=15, d=4)
        assert arith.a1 == Fraction(7, 1)
        assert list(arith) == [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        arith = ArithmeticProgression(n=3, an=15, s=33)
        assert arith.a1 == Fraction(7, 1)
        assert list(arith) == [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

    def test_an(self):
        arith = ArithmeticProgression(n=3, a1=7, d=4)
        assert arith.an == Fraction(15, 1)
        assert list(arith) == [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        arith = ArithmeticProgression(n=3, a1=7, s=33)
        assert arith.an == Fraction(15, 1)
        assert list(arith) == [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

    def test_correct_s(self):
        arith = ArithmeticProgression(a1=3, an=6, s=21)
        assert arith.get_parameters_dict() == {
            "a1": Fraction(3, 1),
            "an": Fraction(6, 1),
            "n": 4,
            "d": Fraction(1, 1),
            "s": Fraction(21, 1),
        }
        assert arith.get_actual_s() == Fraction(18, 1)
        result = list(arith)
        assert result == [
            Fraction(3, 1),
            Fraction(4, 1),
            Fraction(5, 1),
            Fraction(6, 1),
        ]
        assert sum(result) == Fraction(18, 1)

        arith.correct_s = True
        arith.reset_iterator()
        assert arith.get_correction_factor() == Fraction(7, 6)
        result = list(arith)
        assert result == [
            Fraction(7, 2),
            Fraction(14, 3),
            Fraction(35, 6),
            Fraction(7, 1),
        ]
        assert sum(result) == Fraction(21, 1)

    def test_d(self):
        arith = ArithmeticProgression(a1=7, an=15, s=33)
        assert arith.d == Fraction(4, 1)
        assert list(arith) == [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        arith = ArithmeticProgression(a1=7, an=15, n=3)
        assert arith.d == Fraction(4, 1)
        assert list(arith) == [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        arith = ArithmeticProgression(a1=7, n=3, s=33)
        assert arith.d == Fraction(4, 1)
        assert list(arith) == [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        arith = ArithmeticProgression(an=15, n=3, s=33)
        assert arith.d == Fraction(4, 1)
        assert list(arith) == [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        arith = ArithmeticProgression(an=15, s=33)
        with self.assertRaises(DAndSError):
            arith.d = 2

    def test_n(self):
        arith = ArithmeticProgression(an=15, a1=7, d=4)
        assert arith.n == 3
        assert list(arith) == [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        arith = ArithmeticProgression(an=15, a1=7, s=33)
        assert arith.n == 3
        assert list(arith) == [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        arith = ArithmeticProgression(a1=15, n=1, an=30)
        assert arith.d == 0

        with self.assertRaises(AttributeError):
            ArithmeticProgression(a1=15, n=1.4, an=30)

    def test_s(self):
        arith = ArithmeticProgression(a1=7, an=15, d=4)
        assert arith.s == Fraction(33, 1)
        assert list(arith) == [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        arith = ArithmeticProgression(a1=7, an=15, n=3)
        assert arith.s == Fraction(33, 1)
        assert list(arith) == [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        arith = ArithmeticProgression(a1=7, n=3, d=4)
        assert arith.s == Fraction(33, 1)
        assert list(arith) == [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        arith = ArithmeticProgression(an=15, n=3, d=4)
        assert arith.s == Fraction(33, 1)
        assert list(arith) == [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        arith = ArithmeticProgression(an=15, d=4)
        with self.assertRaises(DAndSError):
            arith.s = 30

    def test_get_current_index(self):
        arith = ArithmeticProgression(a1=1, d=2, n=3)
        with self.assertRaises(AttributeError):
            assert arith.get_current_index() is None
        assert next(arith) == Fraction(1, 1)
        assert arith.get_current_index() == 0
        assert next(arith) == Fraction(3, 1)
        assert arith.get_current_index() == 1
        assert next(arith) == Fraction(5, 1)
        assert arith.get_current_index() == 2
        with self.assertRaises(StopIteration):
            next(arith)
        assert arith.get_current_index() == 2

    def test_get_dict(self):
        assert ArithmeticProgression(n=15, a1=1, d=2).get_parameters_dict() == {
            "a1": Fraction(1, 1),
            "an": Fraction(29, 1),
            "n": 15,
            "d": Fraction(2, 1),
            "s": Fraction(225, 1),
        }

    def test_check_args_error(self):
        with self.assertRaises(AttributeError):
            ArithmeticProgression(n=15, a1=1).__next__()

        with self.assertRaises(AttributeError):
            ArithmeticProgression(n=15, a1=1, an=20, d=10).__next__()



from unittest import TestCase

from musurgia.pdf import Pdf
from musurgia.pdf.line import HorizontalLineSegment


class TestPdfSaveState(TestCase):
    def test_save_state(self):
        pdf = Pdf()
        assert (pdf.absolute_x, pdf.absolute_y) == (0, 0)
        pdf.translate_page_margins()
        assert (pdf.absolute_x, pdf.absolute_y) == (10, 10)
        pdf.translate(10, 10)
        assert (pdf.absolute_x, pdf.absolute_y) == (20, 20)
        line = HorizontalLineSegment(length=10, top_margin=10, bottom_margin=20)
        line.draw(pdf)
        assert (pdf.absolute_x, pdf.absolute_y) == (20, 20)

        with pdf.saved_state():
            pdf.translate(10, 10)

        assert (pdf.absolute_x, pdf.absolute_y) == (20, 20)



from musurgia.musurgia_types import check_matrix_data_type
from musurgia.tests.utils_for_tests import PdfTestCase
from musurgia.permutation.limited_permutation import LimitedPermutationOrders


class TestLimitedPermutationOrders(PdfTestCase):
    def test_init_errors(self):
        with self.assertRaises(TypeError):
            LimitedPermutationOrders(main_permutation_order=[3])

        LimitedPermutationOrders(main_permutation_order=(3, 1, 2))

    def test_set_errors(self):
        lt = LimitedPermutationOrders(main_permutation_order=(3, 1, 2))
        with self.assertRaises(TypeError):
            lt.main_permutation_order = [3, 1, 2]

    def test_get_permutation_orders(self):
        lt = LimitedPermutationOrders(main_permutation_order=(3, 1, 4, 2))
        assert lt.get_permutation_orders() == [
            [(3, 1, 4, 2), (4, 3, 2, 1), (2, 4, 1, 3), (1, 2, 3, 4)],
            [(2, 4, 1, 3), (3, 1, 4, 2), (1, 2, 3, 4), (4, 3, 2, 1)],
            [(1, 2, 3, 4), (2, 4, 1, 3), (4, 3, 2, 1), (3, 1, 4, 2)],
            [(4, 3, 2, 1), (1, 2, 3, 4), (3, 1, 4, 2), (2, 4, 1, 3)],
        ]
        assert check_matrix_data_type(lt.get_permutation_orders())
        lt.main_permutation_order = (3, 1, 2)
        assert lt.get_permutation_orders() == [
            [(3, 1, 2), (2, 3, 1), (1, 2, 3)],
            [(1, 2, 3), (3, 1, 2), (2, 3, 1)],
            [(2, 3, 1), (1, 2, 3), (3, 1, 2)],
        ]



from typing import Union, Any, Optional

from musurgia.musurgia_types import (
    check_type,
    PositionType,
    create_error_message,
    LabelPlacement,
    MarginType,
)
from musurgia.pdf.drawobject import HasGetHeightProtocol, PositionedSlave, Master
from musurgia.pdf.margined import Margined
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.positioned import HasPositionsProtocol
from musurgia.pdf.text import AbstractText

__all__ = ["TextLabel"]


class TextLabel(PositionedSlave, AbstractText, Margined):
    def __init__(
        self,
        value: Any,
        master: Optional[Master] = None,
        placement: LabelPlacement = "above",
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self._master: Optional[Master]
        self.master = master
        super().__init__(value=value, *args, **kwargs)  # type: ignore
        self._placement: LabelPlacement
        self.placement = placement

    @property
    def master(self) -> Optional[Master]:
        return self._master

    @master.setter
    def master(self, value: Optional[Master]) -> None:
        self._master = value

    @property
    def placement(self) -> LabelPlacement:
        if not self.master:
            raise AttributeError(
                create_error_message(
                    message="set master first",
                    class_name=self.__class__.__name__,
                    property_name="placement",
                )
            )
        return self._placement

    @placement.setter
    def placement(self, val: LabelPlacement) -> None:
        check_type(
            val,
            "LabelPlacement",
            class_name=self.__class__.__name__,
            property_name="placement",
        )
        self._placement = val


class Labeled(Master, HasPositionsProtocol, HasGetHeightProtocol):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._above_text_labels: list[TextLabel] = []
        self._below_text_labels: list[TextLabel] = []
        self._left_text_labels: list[TextLabel] = []

    def add_text_label(self, label: Union[TextLabel, str], **kwargs: Any) -> TextLabel:
        if not isinstance(label, TextLabel):
            label = TextLabel(label, self, **kwargs)
        else:
            if label.master:
                if label.master != self:
                    raise AttributeError(
                        create_error_message(
                            message=f"label.master {label.master} must be the same as {self}"
                        )
                    )
            else:
                label.master = self
        if label.placement == "above":
            self._above_text_labels.append(label)
        elif label.placement == "below":
            self._below_text_labels.append(label)
        elif label.placement == "left":
            self._left_text_labels.append(label)

        return label

    def add_label(self, label: Union[TextLabel, str], **kwargs: Any) -> TextLabel:
        return self.add_text_label(label, **kwargs)

    def get_above_text_labels(self) -> list[TextLabel]:
        return self._above_text_labels

    def get_below_text_labels(self) -> list[TextLabel]:
        return self._below_text_labels

    def get_left_text_labels(self) -> list[TextLabel]:
        return self._left_text_labels

    def get_text_labels(self) -> list[TextLabel]:
        return (
            self.get_left_text_labels()
            + self.get_above_text_labels()
            + self.get_below_text_labels()
        )

    def draw_above_text_labels(self, pdf: Pdf) -> None:
        if self.get_above_text_labels():
            with pdf.saved_state():
                translate_y = (
                    -self.get_above_text_labels_height()
                    + self.get_above_text_labels()[-1].get_text_height()
                )
                pdf.translate(0, translate_y)
                for text_label in self.get_above_text_labels():
                    text_label.draw(pdf)
                    pdf.translate(0, text_label.get_height())

    def draw_below_text_labels(self, pdf: Pdf) -> None:
        if self.get_below_text_labels():
            with pdf.saved_state():
                pdf.translate(0, self.get_relative_y2() - self.relative_y)
                for text_label in self.get_below_text_labels():
                    pdf.translate(0, text_label.get_height())
                    text_label.draw(pdf)

    def draw_left_text_labels(self, pdf: Pdf) -> None:
        if self.get_left_text_labels():
            with pdf.saved_state():
                pdf.translate(0, -self.get_left_text_labels_height() / 2)
                for text_label in self.get_left_text_labels():
                    pdf.translate(0, text_label.get_height())
                    with pdf.saved_state():
                        pdf.translate(-(text_label.get_width()), 0)
                        text_label.draw(pdf)

    def get_slave_position(self, slave: TextLabel, position: PositionType) -> float:
        check_type(
            position,
            "PositionType",
            class_name=self.__class__.__name__,
            method_name="get_slave_position",
            argument_name="position",
        )
        if not isinstance(slave, TextLabel):
            raise TypeError(
                create_error_message(
                    message=f"slave must be of type TextLabel not {type(slave)}",
                    class_name=self.__class__.__name__,
                    method_name="get_slave_position",
                    argument_name="slave",
                )
            )
        if not slave.master == self:
            raise AttributeError(create_error_message(message="slave hat wrong master"))
        if position == "x":
            return 0
        elif position == "y":
            if slave.placement in ["l", "left"]:
                return (self.get_relative_y2() - self.relative_y) / 2
            return 0

    def get_above_text_labels_height(self) -> float:
        return sum([tl.get_height() for tl in self.get_above_text_labels()])

    def get_left_text_labels_height(self) -> float:
        return sum([tl.get_height() for tl in self.get_left_text_labels()])

    def get_slave_margin(self, slave: Any, margin: MarginType) -> float:
        raise NotImplementedError  # pragma: no cover



from pathlib import Path
from unittest import TestCase

from musurgia.musurgia_exceptions import TimeRulerCannotSetLength
from musurgia.pdf import Pdf
from musurgia.pdf.ruler import TimeRuler
from musurgia.tests.utils_for_tests import PdfTestCase

path = Path(__file__)


class TestTimeRuler(TestCase):
    def setUp(self) -> None:
        self.tr = TimeRuler(duration=200, unit=2, shrink_factor=0.6, mark_line_size=5)

    def test_duration(self):
        assert self.tr.get_length() == 400
        assert self.tr.get_duration() == 200
        assert self.tr.get_unit() == 2

    def test_length_errors(self):
        with self.assertRaises(TimeRulerCannotSetLength):
            TimeRuler(duration=10, length=100)
        with self.assertRaises(AttributeError):
            self.tr.length = 20

    def test_shrink(self):
        assert self.tr.shrink_factor == 0.6
        assert self.tr.mark_line_size == 5
        for i, start_mark_line in enumerate(
            [seg.start_mark_line for seg in self.tr.segments]
        ):
            if i % 10 == 0:
                assert start_mark_line.length == 5
            else:
                assert start_mark_line.length == 3

        self.tr.mark_line_size = 10
        self.tr.shrink_factor = 0.5
        for i, start_mark_line in enumerate(
            [seg.start_mark_line for seg in self.tr.segments]
        ):
            if i % 10 == 0:
                assert start_mark_line.length == 10
            else:
                assert start_mark_line.length == 5
        assert self.tr.segments[-1].end_mark_line.length == 10

        tr = TimeRuler(
            duration=200,
            unit=2,
            label_show_interval=3,
            shrink_factor=0.5,
            mark_line_size=10,
        )
        for i, start_mark_line in enumerate(
            [seg.start_mark_line for seg in tr.segments]
        ):
            if i % 3 == 0:
                assert start_mark_line.length == 10
            else:
                assert start_mark_line.length == 5

    def test_last_segment_end_mark_line(self):
        tr = TimeRuler(duration=200, mark_line_size=10, shrink_factor=0.5)
        assert tr.segments[-1].end_mark_line.length == 10
        tr = TimeRuler(duration=201, mark_line_size=10, shrink_factor=0.5)
        assert tr.segments[-1].end_mark_line.length == 5


class TestTimeRulerPdf(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_time_ruler(self):
        r = TimeRuler(duration=200, unit=2, label_show_interval=10, clock_mode="ms")
        r.change_labels(font_size=6, font_weight="bold")
        r.bottom_margin = 15
        with self.file_path(path, "draw", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            r.clipped_draw(self.pdf)
            self.pdf.write_to_path(pdf_path)



from pathlib import Path

from musurgia.pdf.line import VerticalSegmentedLine
from musurgia.pdf.pdf import Pdf
from musurgia.tests.utils_for_tests import PdfTestCase

path = Path(__file__)


class TestVerticalSegmentedLine(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.vsl = VerticalSegmentedLine(lengths=[10, 15, 20, 25])

    def test_draw(self):
        with self.file_path(path, "draw", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            self.vsl.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_get_width(self):
        self.vsl.segments[1].start_mark_line.length = 5
        actual = self.vsl.get_width()
        expected = 5
        self.assertEqual(expected, actual)

    def test_align_segments(self):
        for index, sg in enumerate(self.vsl.segments):
            sg.start_mark_line.length += index * 5
        self.vsl.segments[-1].end_mark_line.length = (
            self.vsl.segments[-1].start_mark_line.length + 5
        )
        self.vsl.segments[-1].end_mark_line.show = True
        self.vsl._align_segments()
        for seg in self.vsl.segments:
            assert seg.relative_y >= 0
            assert seg.relative_x >= 0
        assert 0 in set([seg.relative_x for seg in self.vsl.segments])



from abc import abstractmethod, ABC
from typing import Optional, Protocol, Any

from musurgia.musurgia_types import ConvertibleToFloat, check_type

__all__ = []  # type: ignore


class HasPositionsProtocol(Protocol):
    @property
    def relative_x(self) -> float: ...

    @property
    def relative_y(self) -> float: ...

    def get_relative_x2(self) -> float: ...

    def get_relative_y2(self) -> float: ...


class AbstractPositioned(ABC):
    """
    An interface for setting and getting DrawObject's position attributes.
    """

    def __init__(
        self,
        relative_x: Optional[ConvertibleToFloat] = None,
        relative_y: Optional[ConvertibleToFloat] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._relative_x: float = 0
        self._relative_y: float = 0
        self.relative_x = relative_x  # type: ignore
        self.relative_y = relative_y  # type: ignore

    @property
    @abstractmethod
    def relative_x(self) -> float:
        """relative_x getter must be provided"""

    @relative_x.setter
    @abstractmethod
    def relative_x(self, val: Optional[ConvertibleToFloat]) -> None:
        """relative_x setter must be provided"""

    @property
    @abstractmethod
    def relative_y(self) -> float:
        """relative_y getter must be provided"""

    @relative_y.setter
    @abstractmethod
    def relative_y(self, val: Optional[ConvertibleToFloat]) -> None:
        """relative_y setter must be provided"""

    def get_positions(self) -> dict[str, float]:
        return {"x": self.relative_x, "y": self.relative_y}

    @property
    def positions(self) -> tuple[float, float]:
        return self.relative_x, self.relative_y

    @positions.setter
    def positions(self, val: tuple[float, float]) -> None:
        self.relative_x, self.relative_y = val


class Positioned(AbstractPositioned):
    def __init__(
        self,
        relative_x: ConvertibleToFloat = 0,
        relative_y: ConvertibleToFloat = 0,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(relative_x=relative_x, relative_y=relative_y, *args, **kwargs)  # type: ignore

    @property
    def relative_x(self) -> float:
        return self._relative_x

    @relative_x.setter
    def relative_x(self, val: ConvertibleToFloat) -> None:
        check_type(
            val,
            "ConvertibleToFloat",
            class_name=self.__class__.__name__,
            property_name="relative_x",
        )
        self._relative_x = float(val)

    @property
    def relative_y(self) -> float:
        return self._relative_y

    @relative_y.setter
    def relative_y(self, val: ConvertibleToFloat) -> None:
        check_type(
            val,
            "ConvertibleToFloat",
            class_name=self.__class__.__name__,
            property_name="relative_y",
        )
        self._relative_y = float(val)



from fractions import Fraction
from typing import Any, Iterator, Literal, Optional, Sequence

from musurgia.musurgia_types import ConvertibleToFraction, convert_to_fraction
from musurgia.quantize import get_quantized_positions


def dToX(
    input_list: Sequence[ConvertibleToFraction],
    first_element: ConvertibleToFraction = 0,
) -> list[Fraction]:
    input = convert_to_fraction_list(input_list)
    output = [convert_to_fraction(first_element)]
    for i in range(len(input)):
        output.append(input[i] + output[i])
    return output


def xToD(input_list: Sequence[ConvertibleToFraction]) -> list[Fraction]:
    input = convert_to_fraction_list(input_list)
    result = []
    for i in range(1, len(input)):
        result.append(input[i] - input[i - 1])
    return result


def flatten(input: list[Any]) -> list[Any]:
    """
    :param input:
    :return:

    >>> flatten([1, 2])
    [1, 2]
    >>> flatten([1, [2, 3], [4, 5, 6]])
    [1, 2, 3, 4, 5, 6]
    >>> flatten([1, [2, 3], [[4, 5, 6], 7, [8, 9]]])
    [1, 2, 3, 4, 5, 6, 7, 8, 9]

    """
    output = []
    for item in input:
        if isinstance(item, list):
            output.extend(flatten(item))
        else:
            output.append(item)
    return output


def convert_to_fraction_list(input: Sequence[ConvertibleToFraction]) -> list[Fraction]:
    return [convert_to_fraction(x) for x in input]


def convert_to_fraction_tuplet(
    input: Sequence[ConvertibleToFraction],
) -> tuple[Fraction, Fraction]:
    if len(input) != 2:
        raise AttributeError
    return (convert_to_fraction(input[0]), convert_to_fraction(input[1]))


class Normalizer:
    def __init__(
        self,
        input_value_range: tuple[ConvertibleToFraction, ConvertibleToFraction],
        normalization_range: tuple[ConvertibleToFraction, ConvertibleToFraction],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._input_value_range: tuple[ConvertibleToFraction, ConvertibleToFraction] = (
            convert_to_fraction_tuplet(input_value_range)
        )
        self._normalization_range: tuple[
            ConvertibleToFraction, ConvertibleToFraction
        ] = convert_to_fraction_tuplet(normalization_range)

    @property
    def normalization_range(
        self,
    ) -> tuple[ConvertibleToFraction, ConvertibleToFraction]:
        return self._normalization_range

    @normalization_range.setter
    def normalization_range(
        self, value: tuple[ConvertibleToFraction, ConvertibleToFraction]
    ) -> None:
        self._normalization_range = convert_to_fraction_tuplet(value)
        if not self._normalization_range[0] <= self._normalization_range[1]:
            raise AttributeError

    @property
    def input_value_range(self) -> tuple[ConvertibleToFraction, ConvertibleToFraction]:
        return self._input_value_range

    @input_value_range.setter
    def input_value_range(
        self, value: tuple[ConvertibleToFraction, ConvertibleToFraction]
    ) -> None:
        self._input_value_range = convert_to_fraction_tuplet(value)
        if not self._input_value_range[0] <= self._input_value_range[1]:
            raise AttributeError

    def get_normalized_value(self, value: ConvertibleToFraction) -> Fraction:
        norm_min, norm_max = self.normalization_range
        value_min, value_max = self.input_value_range
        normalized_value = norm_min + (norm_max - norm_min) * (
            (value - value_min) / (value_max - value_min)
        )
        return convert_to_fraction(normalized_value)


DirectionValueType = Literal[1, -1]


class RelativeValueGenerator:
    def __init__(
        self,
        value_range: tuple[ConvertibleToFraction, ConvertibleToFraction],
        directions: list[DirectionValueType],
        proportions: list[ConvertibleToFraction],
        value_grid: Optional[ConvertibleToFraction] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._calculated_values: list[Fraction]
        self._value_range: tuple[ConvertibleToFraction, ConvertibleToFraction] = (
            convert_to_fraction_tuplet(value_range)
        )
        self._directions: list[DirectionValueType] = directions
        self._proportions: list[Fraction] = convert_to_fraction_list(proportions)
        self._value_grid: Optional[ConvertibleToFraction] = (
            None if not value_grid else convert_to_fraction(value_grid)
        )

        self._calculate_intervals()
        self._normalizer: Normalizer
        self._set_normalizer()
        self._calculate_result_values()

    def _calculate_intervals(self) -> None:
        self._intervals = [
            proportion * direction
            for proportion, direction in zip(self.proportions, self.directions)
        ]

    def _get_input_values(self) -> list[Fraction]:
        return dToX(self._intervals)

    def _set_normalizer(self) -> None:
        input_values = self._get_input_values()
        self._normalizer = Normalizer(
            input_value_range=(min(input_values), max(input_values)),
            normalization_range=self.value_range,
        )

    def _calculate_result_values(self) -> None:
        normalized_values = [
            self._normalizer.get_normalized_value(value)
            for value in self._get_input_values()
        ]
        if self.value_grid:
            self._calculated_values = get_quantized_positions(
                normalized_values, grid_size=self.value_grid
            )
        else:
            self._calculated_values = normalized_values

    @property
    def value_range(self) -> tuple[ConvertibleToFraction, ConvertibleToFraction]:
        return self._value_range

    @value_range.setter
    def value_range(
        self, value: tuple[ConvertibleToFraction, ConvertibleToFraction]
    ) -> None:
        self._value_range = convert_to_fraction_tuplet(value)
        self._set_normalizer()
        self._calculate_result_values()

    @property
    def directions(self) -> list[DirectionValueType]:
        return self._directions

    @directions.setter
    def directions(self, value: list[DirectionValueType]) -> None:
        self._directions = value
        self._calculate_intervals()
        self._calculate_result_values()

    @property
    def proportions(self) -> list[Fraction]:
        return self._proportions

    @proportions.setter
    def proportions(self, value: Sequence[ConvertibleToFraction]) -> None:
        self._proportions = convert_to_fraction_list(value)
        self._calculate_intervals()
        self._set_normalizer()
        self._calculate_result_values()

    @property
    def value_grid(self) -> Optional[ConvertibleToFraction]:
        return self._value_grid

    @value_grid.setter
    def value_grid(self, value: Optional[ConvertibleToFraction]) -> None:
        self._value_grid = value
        self._calculate_result_values()

    def get_values(self) -> list[Fraction]:
        return self._calculated_values

    def __iter__(self) -> Iterator[Fraction]:
        for val in self.get_values():
            yield val



import copy
from typing import cast, Any, Union

from verysimpletree.tree import Tree, T

from musurgia.musurgia_exceptions import SegmentedLineSegmentHasMarginsError
from musurgia.pdf import DrawObjectRow, Pdf, DrawObjectColumn, HorizontalLineSegment
from musurgia.pdf.line import LineSegment
from musurgia.trees.valuedtree import ValuedTree


__all__ = ["ValuedTreeNodeSegment", "GraphicChildrenSegmentedLine", "GraphicTree"]


class ValuedTreeNodeSegment(HorizontalLineSegment):
    def __init__(
        self, node_value: float, unit: int = 1, *args: Any, **kwargs: Any
    ) -> None:
        super().__init__(length=node_value * unit, *args, **kwargs)  # type: ignore
        self._node_value = node_value
        self._unit: int
        self.unit = unit

    def _update_length(self) -> None:
        self.straight_line.length = self.get_node_value() * self.unit

    @property  # type: ignore
    def length(self) -> float:
        return super().length

    @property
    def unit(self) -> int:
        return self._unit

    @unit.setter
    def unit(self, value: int) -> None:
        self._unit = value
        self._update_length()

    def get_node_value(self) -> float:
        return self._node_value

    def set_node_value(self, val: float) -> None:
        self._node_value = val
        self._update_length()


class GraphicChildrenSegmentedLine(DrawObjectRow):
    def __init__(
        self, graphic_children: list["GraphicTree"], *args: Any, **kwargs: Any
    ):
        super().__init__(*args, **kwargs)
        self._children = graphic_children
        self._create_segments()

    def _create_segments(self) -> None:
        segments = [copy.deepcopy(node.get_segment()) for node in self._children]
        for seg in segments:
            self.add_draw_object(seg)

    def _check_segment_margins(self) -> None:
        for seg in self.segments:
            if seg.margins != (0, 0, 0, 0):
                raise SegmentedLineSegmentHasMarginsError()

    @property
    def segments(self) -> list[LineSegment]:
        return cast(list[LineSegment], self.get_draw_objects())

    def set_straight_line_relative_y(self, val: Union[int, float]) -> None:
        delta = val - self.segments[0].straight_line.relative_y
        self.relative_y += delta

    def _align_segments(self) -> None:
        reference_segment = max(self.segments, key=lambda seg: seg.get_height())
        for segment in self.segments:
            if segment != reference_segment:
                segment.set_straight_line_relative_y(
                    reference_segment.straight_line.relative_y
                )

    def draw(self, pdf: Pdf) -> None:
        self._check_segment_margins()
        self._align_segments()
        self.segments[-1].end_mark_line.show = True
        super().draw(pdf)


class GraphicTree(Tree[Any]):
    def __init__(
        self,
        valued_tree: ValuedTree,
        distance: Union[int, float] = 5,
        unit: int = 1,
        mark_line_length: Union[int, float] = 6,
        shrink_factor: Union[int, float] = 0.7,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._valued_tree: ValuedTree = valued_tree
        self._distance: Union[int, float]
        self._unit: int
        self._mark_line_length: Union[int, float]
        self._shrink_factor: Union[int, float]

        self.distance = distance
        self.unit = unit
        self.mark_line_length = mark_line_length
        self.shrink_factor = shrink_factor

        self._segment: ValuedTreeNodeSegment
        self._graphic: DrawObjectColumn
        self._populate_tree()

    def _check_child_to_be_added(self: T, child: T) -> bool:
        if not isinstance(child, GraphicTree):
            raise TypeError
        else:
            return True

    def _create_graphic(self) -> None:
        self._graphic = DrawObjectColumn()
        self._graphic.add_draw_object(self.get_segment())
        if not self.is_leaf:
            second_row = DrawObjectRow()
            for ch in self.get_children():
                ch._create_graphic()
                second_row.add_draw_object(ch.get_graphic())
            self._graphic.add_draw_object(second_row)

    def _populate_tree(self) -> None:
        self._segment = ValuedTreeNodeSegment(
            node_value=float(self.get_valued_tree().get_value()), unit=self.unit
        )
        self._segment.start_mark_line.length = self.mark_line_length
        self._segment.end_mark_line.length = self.mark_line_length
        for ch in self.get_valued_tree().get_children():
            self.add_child(
                GraphicTree(
                    ch,
                    distance=self.distance,
                    unit=self.unit,
                    mark_line_length=self.mark_line_length * self.shrink_factor,
                )
            )

    def _show_last_mark_lines(self) -> None:
        for node in self.traverse():
            if node.is_last_child:
                node.get_segment().end_mark_line.show = True

    def _update_mark_line_length(self) -> None:
        if self.is_first_child:
            self.get_segment().start_mark_line.length = self.mark_line_length
        else:
            self.get_segment().start_mark_line.length = (
                self.mark_line_length * self.shrink_factor
            )
        if self.is_last_child:
            self.get_segment().end_mark_line.length = self.mark_line_length

    def _update_mark_line_lengths(self) -> None:
        for node in self.traverse():
            node._update_mark_line_length()

    #     # def _align_layer_segments(self):
    #     #     for layer_number in range(1, self.get_number_of_layers() + 1):
    #     #         current_layer = self.get_layer(layer_number)
    #     #         segments = [l.get_segment() for l in current_layer]
    #     #         current_layer_largeste_mark_line = get_largest_mark_line(segments)
    #     #         for seg in segments:
    #     #             max_mark_line_length = max([seg.start_mark_line.length, seg.end_mark_line.length])
    #     #             dy = (current_layer_largeste_mark_line.length - max_mark_line_length) / 2
    #     #             seg.relative_y += dy
    def _update_distance(self) -> None:
        for node in self.traverse():
            if not node.is_leaf:
                node._graphic.get_draw_objects()[1].relative_y += node.distance

    def _finalize(self) -> None:
        self._show_last_mark_lines()
        self._update_mark_line_lengths()
        self._update_distance()

    @property
    def distance(self) -> Union[int, float]:
        return self._distance

    @distance.setter
    def distance(self, val: Union[int, float]) -> None:
        self._distance = val

    @property
    def mark_line_length(self) -> Union[int, float]:
        return self._mark_line_length

    @mark_line_length.setter
    def mark_line_length(self, val: Union[int, float]) -> None:
        self._mark_line_length = round(val, 5)

    @property
    def unit(self) -> int:
        return self._unit

    @unit.setter
    def unit(self, value: int) -> None:
        self._unit = value
        for node in self.traverse():
            if node != self:
                node.unit = value
        try:
            self._segment.unit = self.unit
        except AttributeError:
            pass

    @property
    def shrink_factor(self) -> Union[int, float]:
        return self._shrink_factor

    @shrink_factor.setter
    def shrink_factor(self, val: Union[int, float]) -> None:
        self._shrink_factor = val

    def create_layer_graphic(self, layer_number: int) -> GraphicChildrenSegmentedLine:
        self._update_mark_line_lengths()
        children = self.get_layer(layer_number)
        return GraphicChildrenSegmentedLine(graphic_children=children)

    def get_segment(self) -> ValuedTreeNodeSegment:
        return self._segment

    def get_valued_tree(self) -> ValuedTree:
        return self._valued_tree

    def get_graphic(self) -> DrawObjectColumn:
        for node in self.traverse():
            node._create_graphic()
        self._finalize()
        return self._graphic

    def set_all_distances(self, val: Union[int, float]) -> None:
        for node in self.traverse():
            node.distance = val

    def draw(self, pdf: Pdf) -> None:
        self.get_graphic().draw(pdf)



from typing import Optional, Any, TypeVar

from musurgia.musurgia_exceptions import (
    MatrixIsEmptyError,
    SquareMatrixDataError,
    PermutationOrderMatrixDataError,
    MatrixIndexEndOfMatrixError,
    MatrixIndexEndOfRowError,
    MatrixIndexControllerReadingDirectionError,
)
from musurgia.musurgia_types import (
    NonNegativeInteger,
    check_type,
    PositiveInteger,
    MatrixData,
    MatrixIndex,
    check_matrix_index_values,
    check_permutation_order_values,
    create_error_message,
    PermutationOrder,
    MatrixTransposeMode,
    MatrixReadingDirection,
)
from musurgia.permutation.limited_permutation import LimitedPermutationOrders


class SimpleMatrix:
    T = TypeVar("T", bound="SimpleMatrix")
    """
    SimpleMatrix is the core class class for representing a list of lists.
    """

    def __init__(
        self, matrix_data: Optional[MatrixData] = None, *args: Any, **kwargs: Any
    ) -> None:
        self._matrix_data: MatrixData = []
        if matrix_data is None:
            matrix_data = []
        self.matrix_data = matrix_data

    @property
    def is_empty(self) -> bool:
        if not self.matrix_data:
            return True
        return False

    @property
    def matrix_data(self) -> MatrixData:
        return self._matrix_data

    @matrix_data.setter
    def matrix_data(self, val: MatrixData) -> None:
        check_type(
            val,
            "MatrixData",
            class_name=self.__class__.__name__,
            property_name="matrix_data",
        )
        self._matrix_data = val

    def get_row_size(self) -> NonNegativeInteger:
        try:
            return len(self.matrix_data[0])
        except IndexError:
            return 0

    def get_column_size(self) -> NonNegativeInteger:
        return len(self.matrix_data)

    def get_row(self, row_number: PositiveInteger) -> list[Any]:
        check_type(
            v=row_number,
            t="PositiveInteger",
            class_name=self.__class__.__name__,
            method_name="get_column",
            argument_name="row_number",
        )
        if self.is_empty:
            raise MatrixIsEmptyError()
        if row_number > self.get_column_size():
            raise ValueError(
                f"{self.__class__.__name__}:get_column:column_number must be less than or equal to {self.get_column_size()}"
            )
        return self.matrix_data[row_number - 1]

    def get_column(self, column_number: PositiveInteger) -> list[Any]:
        check_type(
            v=column_number,
            t="PositiveInteger",
            class_name=self.__class__.__name__,
            method_name="get_column",
            argument_name="column_number",
        )
        if self.is_empty:
            raise MatrixIsEmptyError()
        if column_number > self.get_row_size():
            raise ValueError(
                f"{self.__class__.__name__}:get_column:column_number must be less than or equal to {self.get_row_size()}"
            )
        return self.get_transposed_matrix().matrix_data[column_number - 1]

    def get_element(self, element_index: MatrixIndex) -> Any:
        if self.is_empty:
            raise MatrixIsEmptyError()
        check_type(
            v=element_index,
            t="MatrixIndex",
            class_name=self.__class__.__name__,
            method_name="get_element",
            argument_name="element_index",
        )
        check_matrix_index_values(
            element_index,
            number_of_rows=self.get_column_size(),
            number_of_columns=self.get_row_size(),
        )
        return self.matrix_data[element_index[0] - 1][element_index[1] - 1]

    def get_transposed_matrix(
        matrix: "T", mode: MatrixTransposeMode = "regular"
    ) -> "T":
        check_type(
            v=mode,
            t="MatrixTransposeMode",
            class_name="MatrixTransposition",
            method_name="get_transposed_matrix",
            argument_name="mode",
        )
        controller = MatrixIndexController(
            number_of_rows=matrix.get_column_size(),
            number_of_columns=matrix.get_row_size(),
        )
        if mode == "regular":
            controller.reading_direction = "vertical"
        elif mode == "diagonal":
            controller.reading_direction = "diagonal"
        matrix_data = []
        for c in range(controller.number_of_columns):
            matrix_data.append(
                [
                    matrix.get_element(next(controller))
                    for _ in range(controller.number_of_rows)
                ]
            )
        return matrix.__class__(matrix_data=matrix_data)


class Matrix(SimpleMatrix):
    T = TypeVar("T", bound="Matrix")

    def add_row(self, row: list[Any]) -> None:
        check_type(
            v=row,
            t=list,
            class_name=self.__class__.__name__,
            method_name="add_row",
            argument_name="row",
        )
        if self.get_row_size() and len(row) != self.get_row_size():
            raise ValueError(
                f"{self.__class__.__name__}:add_row:row must be of size ({self.get_row_size()})"
            )
        self.matrix_data.append(row)

    def remove_row(self, row_number: NonNegativeInteger) -> list[Any]:
        if self.is_empty:
            raise MatrixIsEmptyError()
        check_type(
            v=row_number,
            t="PositiveInteger",
            class_name=self.__class__.__name__,
            method_name="remove_row",
            argument_name="row_number",
        )
        if row_number > self.get_column_size():
            raise ValueError(
                f"{self.__class__.__name__}:remove_row:row_number must be less than or equal to {self.get_column_size()}"
            )
        return self.matrix_data.pop(row_number - 1)


class SquareMatrix(SimpleMatrix):
    T = TypeVar("T", bound="SimpleMatrix")

    def __init__(self, matrix_data: MatrixData, *args: Any, **kwargs: Any) -> None:
        if matrix_data is None or matrix_data == []:
            raise SquareMatrixDataError(
                "SquareMatrix.matrix_data cannot be empty or None"
            )
        super().__init__(matrix_data=matrix_data, *args, **kwargs)  # type: ignore
        if self.get_row_size() != self.get_column_size():
            raise SquareMatrixDataError(
                f"SquareMatrix.matrix_data: row size {self.get_row_size()} must be equal to column size {self.get_column_size()}"
            )

    def get_size(self) -> NonNegativeInteger:
        return self.get_column_size()


class PermutationOrderMatrixGenerator:
    def __init__(self, main_permutation_order: PermutationOrder):
        self._lp = LimitedPermutationOrders(main_permutation_order)

    @property
    def main_permutation_order(self) -> PermutationOrder:
        return self._lp.main_permutation_order

    @main_permutation_order.setter
    def main_permutation_order(self, value: PermutationOrder) -> None:
        self._lp.main_permutation_order = value

    def generate_permutation_order_matrix(self) -> "PermutationOrderMatrix":
        return PermutationOrderMatrix(matrix_data=self._lp.get_permutation_orders())


class PermutationOrderMatrix(SquareMatrix):
    T = TypeVar("T", bound="SquareMatrix")

    @property
    def matrix_data(self) -> MatrixData:
        return super().matrix_data

    @matrix_data.setter
    def matrix_data(self, val: MatrixData) -> None:
        check_type(
            val,
            "MatrixData",
            class_name=self.__class__.__name__,
            property_name="matrix_data",
        )
        self._check_matrix_data_permutation_orders(val)
        self._matrix_data = val

    def _check_matrix_data_permutation_orders(self, matrix_data: MatrixData) -> bool:
        size = None
        for permutation_order in [_ for row in matrix_data for _ in row]:
            try:
                check_type(
                    v=permutation_order,
                    t="PermutationOrder",
                    class_name=self.__class__.__name__,
                    property_name="matrix_data",
                )
            except TypeError as err:
                raise PermutationOrderMatrixDataError(err)
            if size is None:
                size = len(permutation_order)
            else:
                try:
                    check_permutation_order_values(
                        permutation_order=permutation_order, size=size
                    )
                except ValueError as err:
                    raise PermutationOrderMatrixDataError(
                        create_error_message(
                            class_name=self.__class__.__name__,
                            property_name="matrix_data",
                            message=str(err),
                        )
                    )
        return True


class MatrixIndexController:
    def __init__(
        self,
        number_of_rows: NonNegativeInteger,
        number_of_columns: NonNegativeInteger,
        first_index: MatrixIndex = (1, 1),
        reading_direction: MatrixReadingDirection = "horizontal",
    ):
        self._number_of_rows: NonNegativeInteger
        self._number_of_columns: NonNegativeInteger
        self._reading_direction: MatrixReadingDirection
        self._first_index: MatrixIndex
        self._flatten_index = 0

        self.number_of_rows = number_of_rows
        self.number_of_columns = number_of_columns
        self.reading_direction = reading_direction

        self.first_index = first_index

    def _convert_flatten_index_to_index_horizontal(
        self, flatten_index: PositiveInteger
    ) -> MatrixIndex:
        r = flatten_index // self.number_of_columns + 1
        c = flatten_index % self.number_of_columns + 1
        return r, c

    def _convert_flatten_index_to_index_diagonal(
        self, flatten_index: NonNegativeInteger
    ) -> MatrixIndex:
        r = (flatten_index + 1) % self.number_of_rows
        if r == 0:
            r = self.number_of_rows
        c = (r + flatten_index // self.number_of_rows) % self.number_of_columns
        if c == 0:
            c = self.number_of_columns
        return r, c

    def _convert_flatten_index_to_index_vertical(
        self, flatten_index: NonNegativeInteger
    ) -> MatrixIndex:
        c = flatten_index // self.number_of_rows + 1
        r = flatten_index % self.number_of_rows + 1
        return r, c

    def _convert_flatten_index_to_index(
        self, flatten_index: NonNegativeInteger
    ) -> MatrixIndex:
        check_type(
            flatten_index,
            "NonNegativeInteger",
            class_name=self.__class__.__name__,
            method_name="_convert_flatten_index_to_index",
            argument_name="flatten_index",
        )
        if flatten_index >= self.number_of_rows * self.number_of_columns:
            raise MatrixIndexEndOfMatrixError
        if self.reading_direction == "horizontal":
            return self._convert_flatten_index_to_index_horizontal(flatten_index)

        elif self.reading_direction == "diagonal":
            return self._convert_flatten_index_to_index_diagonal(flatten_index)

        elif self.reading_direction == "vertical":
            return self._convert_flatten_index_to_index_vertical(flatten_index)

    def _convert_index_to_flatten_index(self, index: MatrixIndex) -> PositiveInteger:
        r = index[0]
        c = index[1]
        if self.reading_direction == "horizontal":
            flatten_index = c + (r - 1) * self.number_of_columns
            flatten_index -= 1
            return flatten_index
        elif self.reading_direction == "diagonal":
            flatten_index = (c - 1) * self.number_of_rows + (r - 1) * (
                (self.number_of_rows * (self.number_of_columns - 1)) + 1
            )
            flatten_index %= self.number_of_columns * self.number_of_rows
            return flatten_index
        elif self.reading_direction == "vertical":
            flatten_index = (r - 1) + (c - 1) * self.number_of_rows
            return flatten_index

    def _get_next_index(self) -> MatrixIndex:
        return self._convert_flatten_index_to_index(self._flatten_index)

    @property
    def number_of_rows(self) -> NonNegativeInteger:
        return self._number_of_rows

    @number_of_rows.setter
    def number_of_rows(self, value: NonNegativeInteger) -> None:
        check_type(
            value,
            "NonNegativeInteger",
            class_name=self.__class__.__name__,
            property_name="number_of_rows",
        )
        self._number_of_rows = value
        self.reset()

    @property
    def number_of_columns(self) -> NonNegativeInteger:
        return self._number_of_columns

    @number_of_columns.setter
    def number_of_columns(self, value: NonNegativeInteger) -> None:
        check_type(
            value,
            "NonNegativeInteger",
            class_name=self.__class__.__name__,
            property_name="number_of_columns",
        )
        self._number_of_columns = value
        self.reset()

    @property
    def reading_direction(self) -> MatrixReadingDirection:
        return self._reading_direction

    @reading_direction.setter
    def reading_direction(self, value: MatrixReadingDirection) -> None:
        check_type(
            value,
            "MatrixReadingDirection",
            class_name=self.__class__.__name__,
            property_name="reading_direction",
        )
        self._reading_direction = value
        self.reset()

    @property
    def first_index(self) -> Optional[MatrixIndex]:
        return self._first_index

    @first_index.setter
    def first_index(self, value: MatrixIndex) -> None:
        check_type(
            value,
            "MatrixIndex",
            class_name=self.__class__.__name__,
            property_name="index",
        )
        check_matrix_index_values(value, self.number_of_rows, self.number_of_columns)
        self._first_index = value
        self._flatten_index = self._convert_index_to_flatten_index(value)

    def get_next_in_row(self) -> MatrixIndex:
        if self.reading_direction != "horizontal":
            raise MatrixIndexControllerReadingDirectionError
        next_index = self._get_next_index()
        if next_index != self.first_index and next_index[1] == 1:
            raise MatrixIndexEndOfRowError
        return self.__next__()

    def get_next_flatten_index(self) -> int:
        return self._flatten_index

    def reset(self) -> None:
        try:
            self.first_index = self._first_index
        except AttributeError:
            pass

    def __iter__(self) -> "MatrixIndexController":
        return self

    def __next__(self) -> MatrixIndex:
        output = self._get_next_index()
        self._flatten_index += 1
        return output



from abc import ABC, abstractmethod
from math import ceil
from typing import Optional, Protocol, Any

from musurgia.musurgia_exceptions import (
    PdfAttributeError,
    RelativePositionNotSettableError,
    MarginNotSettableError,
)
from musurgia.musurgia_types import (
    create_error_message,
    check_type,
    MusurgiaTypeError,
    PositionType,
    MarginType,
)
from musurgia.pdf.margined import AbstractMargined, Margined
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.positioned import AbstractPositioned, Positioned

__all__ = ["ClippingArea"]


class HasGetHeightProtocol(Protocol):
    def get_height(self) -> float: ...


class HasShowProtocol(Protocol):
    @property
    def show(self) -> bool: ...


class DrawObject(AbstractPositioned, AbstractMargined, ABC):
    def __init__(self, show: bool = True, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._show: bool
        self.show = show
        self._clipping_area = ClippingArea(pdf=None, draw_object=self)

    def get_clipping_area(self) -> "ClippingArea":
        return self._clipping_area

    @property
    def show(self) -> bool:
        return self._show

    @show.setter
    def show(self, val: bool) -> None:
        check_type(val, bool, class_name=self.__class__.__name__, property_name="show")
        self._show = val

    @abstractmethod
    def get_relative_x2(self) -> float:
        """this property is needed to get relative_x2"""

    @abstractmethod
    def get_relative_y2(self) -> float:
        """this property is needed to get relative_y2"""

    def get_end_positions(self) -> tuple[float, float]:
        return self.get_relative_x2(), self.get_relative_y2()

    def get_height(self) -> float:
        return (
            self.top_margin
            + self.get_relative_y2()
            - self.relative_y
            + self.bottom_margin
        )

    def get_width(self) -> float:
        return (
            self.left_margin
            + self.get_relative_x2()
            - self.relative_x
            + self.right_margin
        )

    @abstractmethod
    def draw(self, pdf: Pdf) -> None:
        """this property is needed draw the DrawObject to pdf"""

    def clipped_draw(self, pdf: Pdf) -> None:
        self.get_clipping_area().pdf = pdf
        self.get_clipping_area().draw()


class Master(ABC):
    @abstractmethod
    def get_slave_position(self, slave: Any, position: PositionType) -> float:
        """get_slave_position must be provided"""

    @abstractmethod
    def get_slave_margin(self, slave: Any, margin: MarginType) -> float:
        """get_slave_margin must be provided"""


class HasMasterProtocol(Protocol):
    @property
    def master(self) -> Optional["Master"]: ...


class PositionedSlave(AbstractPositioned, HasMasterProtocol):
    @property
    def relative_x(self) -> float:
        if not self.master:
            raise AttributeError(
                create_error_message(
                    message="set master first",
                    class_name=self.__class__.__name__,
                    property_name="relative_x",
                )
            )
        return self.master.get_slave_position(slave=self, position="x")

    @relative_x.setter
    def relative_x(self, val: Optional[float]) -> None:
        if val is not None:
            raise RelativePositionNotSettableError()

    @property
    def relative_y(self) -> float:
        if not self.master:
            raise AttributeError(
                create_error_message(
                    message="set master first",
                    class_name=self.__class__.__name__,
                    property_name="relative_y",
                )
            )
        return self.master.get_slave_position(slave=self, position="y")

    @relative_y.setter
    def relative_y(self, val: Optional[float]) -> None:
        if val is not None:
            raise RelativePositionNotSettableError()


class MarginedSlave(AbstractMargined, HasMasterProtocol):
    @property
    def left_margin(self) -> float:
        if not self.master:
            raise AttributeError(
                create_error_message(
                    message="set master first",
                    class_name=self.__class__.__name__,
                    property_name="left_margin",
                )
            )
        return self.master.get_slave_margin(slave=self, margin="left")

    @left_margin.setter
    def left_margin(self, val: Optional[float]) -> None:
        if val is not None:
            raise MarginNotSettableError()

    @property
    def top_margin(self) -> float:
        if not self.master:
            raise AttributeError(
                create_error_message(
                    message="set master first",
                    class_name=self.__class__.__name__,
                    property_name="top_margin",
                )
            )
        return self.master.get_slave_margin(slave=self, margin="top")

    @top_margin.setter
    def top_margin(self, val: Optional[float]) -> None:
        if val is not None:
            raise MarginNotSettableError()

    @property
    def right_margin(self) -> float:
        if not self.master:
            raise AttributeError(
                create_error_message(
                    message="set master first",
                    class_name=self.__class__.__name__,
                    property_name="right_margin",
                )
            )
        return self.master.get_slave_margin(slave=self, margin="right")

    @right_margin.setter
    def right_margin(self, val: Optional[float]) -> None:
        if val is not None:
            raise MarginNotSettableError()

    @property
    def bottom_margin(self) -> float:
        if not self.master:
            raise AttributeError(
                create_error_message(
                    message="set master first",
                    class_name=self.__class__.__name__,
                    property_name="bottom_margin",
                )
            )
        return self.master.get_slave_margin(slave=self, margin="bottom")

    @bottom_margin.setter
    def bottom_margin(self, val: Optional[float]) -> None:
        if val is not None:
            raise MarginNotSettableError()


class MasterDrawObject(Master, DrawObject, Positioned, Margined, ABC):
    pass


class SlaveDrawObject(DrawObject, PositionedSlave, MarginedSlave, ABC):
    def __init__(
        self,
        master: Optional[MasterDrawObject] = None,
        simple_name: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ):
        self._master: Optional[MasterDrawObject]
        self.master = master
        super().__init__(*args, **kwargs)

        self._simple_name: Optional[str]
        self.simple_name = simple_name

    @property
    def master(self) -> Optional[MasterDrawObject]:
        return self._master

    @master.setter
    def master(self, val: Optional[MasterDrawObject]) -> None:
        if val is not None and not isinstance(val, MasterDrawObject):
            raise MusurgiaTypeError(
                message=f"master.value must be of type {MasterDrawObject} not {type(val)}",
                class_name=self.__class__.__name__,
                property_name="master",
            )
        self._master = val


class ClippingArea:
    def __init__(
        self,
        pdf: Optional[Pdf],
        draw_object: DrawObject,
        left_margin: float = 0,
        right_margin: float = 0,
        top_margin: float = 0,
    ):
        self.pdf: Optional[Pdf] = pdf
        self.draw_object: DrawObject = draw_object
        self.left_margin: float = left_margin
        self.right_margin: float = right_margin
        self.top_margin: float = top_margin

    # private methods
    def _add_page(self) -> None:
        if not self.pdf:
            raise PdfAttributeError(
                self._get_pdf_not_exists_message("_add_page")
            )  # pragma: no cover
        self.pdf.add_page()
        self._prepare_page()

    def _get_pdf_not_exists_message(self, method_name: str) -> str:
        return create_error_message(
            message="pdf must be set first",
            class_name=self.__class__.__name__,
            method_name=method_name,
        )

    def _draw_with_clip(self, index: int) -> None:
        if not self.pdf:
            raise PdfAttributeError(
                self._get_pdf_not_exists_message("_draw_with_clip")
            )  # pragma: no cover
        with self.pdf.saved_state():
            self.pdf.clip_rect(
                -1, -5, self.get_row_width() + 1.14, self.get_row_height()
            )
            self.pdf.translate(index * -self.get_row_width(), 0)
            self.draw_object.draw(self.pdf)

    def _prepare_page(self) -> None:
        if not self.pdf:
            raise PdfAttributeError(
                self._get_pdf_not_exists_message("_prepare_page")
            )  # pragma: no cover
        self.pdf.translate_page_margins()
        self.pdf.translate(self.left_margin, self.top_margin)

    # public methods
    def draw(self) -> None:
        if not self.pdf:
            raise PdfAttributeError(self._get_pdf_not_exists_message("draw"))
        self.pdf.translate(self.left_margin, self.top_margin)
        for index in range(self.get_number_of_rows()):
            if index != 0:
                self.pdf.translate(0, self.draw_object.get_height())
            if self.pdf.absolute_y > self.pdf.h - self.pdf.b_margin:
                self._add_page()
            self._draw_with_clip(index)

    def get_number_of_rows(self) -> int:
        return int(ceil(self.draw_object.get_width() / self.get_row_width()))

    def get_row_height(self) -> float:
        return self.draw_object.get_height()

    def get_row_width(self) -> float:
        if not self.pdf:
            raise PdfAttributeError(self._get_pdf_not_exists_message("get_row_width"))
        return (
            self.pdf.w
            - self.pdf.l_margin
            - self.pdf.r_margin
            - self.left_margin
            - self.right_margin
        )



from fractions import Fraction
from unittest import TestCase

from musurgia.musurgia_types import (
    check_type,
    MusurgiaTypeError,
    check_matrix_index_values,
    check_permutation_order_values,
    create_error_message,
    convert_to_fraction,
    LiteralCheckGenerator,
    HorizontalVertical,
)


class TestCreateErrorMessage(TestCase):
    def test_check_create_message_errors(self):
        with self.assertRaises(AttributeError):
            create_error_message(v=3, t=list, function_name="function_name")
        with self.assertRaises(AttributeError):
            create_error_message(
                v=3, t=list, function_name="function_name", class_name="ClassName"
            )
        with self.assertRaises(AttributeError):
            create_error_message(
                v=3,
                t=list,
                function_name="function_name",
                property_name="property_name",
            )
        with self.assertRaises(AttributeError):
            create_error_message(
                v=3, t=list, function_name="function_name", method_name="method_name"
            )
        with self.assertRaises(AttributeError):
            create_error_message(v=3, t=list, class_name="ClassName")
        with self.assertRaises(AttributeError):
            create_error_message(
                v=3, t=list, class_name="ClassName", argument_name="argument_name"
            )
        with self.assertRaises(AttributeError):
            create_error_message(
                v=3, t=list, class_name="ClassName", method_name="method_name"
            )
        with self.assertRaises(AttributeError):
            create_error_message(
                v=3,
                t=list,
                class_name="ClassName",
                property_name="property_name",
                argument_name="argument_name",
            )
        with self.assertRaises(AttributeError):
            create_error_message(
                v=3,
                t=list,
                class_name="ClassName",
                class_attribute_name="class_attribute_name",
                property_name="property_name",
            )
        with self.assertRaises(AttributeError):
            create_error_message(
                v=3,
                t=list,
                class_name="ClassName",
                class_attribute_name="class_attribute_name",
                method_name="method_name",
            )
        with self.assertRaises(AttributeError):
            create_error_message(
                v=3,
                t=list,
                class_name="ClassName",
                class_attribute_name="class_attribute_name",
                argument_name="argument_name",
            )
        with self.assertRaises(AttributeError):
            create_error_message(
                v=3, t=list, class_attribute_name="class_attribute_name"
            )
        with self.assertRaises(AttributeError):
            create_error_message(v=3, t=list, method_name="method_name")
        with self.assertRaises(AttributeError):
            create_error_message(message="message", method_name="method_name")
        with self.assertRaises(AttributeError):
            create_error_message(v=3, t=list, property_name="property_name")
        with self.assertRaises(AttributeError):
            create_error_message(v=3, t=list, property_name="argument_name")
        with self.assertRaises(AttributeError):
            create_error_message(v=3, t=list, message="message")
        with self.assertRaises(AttributeError):
            create_error_message(v=3, message="message")
        with self.assertRaises(AttributeError):
            create_error_message(t=list, message="message")
        with self.assertRaises(AttributeError):
            create_error_message()

    def test_check_create_message(self):
        assert (
            create_error_message(v=3, t=list) == "Value 3 must be of type list not int"
        )
        assert (
            create_error_message(
                v=3,
                t=list,
                function_name="function_name",
                argument_name="argument_name",
            )
            == "function_name:argument_name: Value 3 must be of type list not int"
        )
        assert (
            create_error_message(
                v=3,
                t=list,
                class_name="ClassName",
                method_name="method_name",
                argument_name="argument_name",
            )
            == "ClassName.method_name:argument_name: Value 3 must be of type list not int"
        )
        assert (
            create_error_message(
                v=3, t=list, class_name="ClassName", property_name="property_name"
            )
            == "ClassName.property_name: Value 3 must be of type list not int"
        )
        assert (
            create_error_message(
                class_name="ClassName",
                property_name="property_name",
                message=f"Value {3} must be greater than 5",
            )
            == "ClassName.property_name: Value 3 must be greater than 5"
        )


class TestCheckTypeFunction(TestCase):
    def test_matrix_data(self):
        assert check_type([], "MatrixData")
        assert check_type([[1, 2, 3]], "MatrixData")
        assert check_type([[1, 2, 3], [4, 5, 6]], "MatrixData")

        with self.assertRaises(MusurgiaTypeError):
            check_type(0, "MatrixData")
        with self.assertRaises(MusurgiaTypeError):
            check_type([1, 2, 3], "MatrixData")
        with self.assertRaises(MusurgiaTypeError):
            check_type([[1, 2, 3], [1, 2]], "MatrixData")

    def test_matrix_index(self):
        # wrong type
        with self.assertRaises(TypeError):
            check_type([1, 2], "MatrixIndex")
        with self.assertRaises(TypeError):
            check_type((-1, 2), "MatrixIndex")
        # wrong lengths
        with self.assertRaises(TypeError):
            check_type((1, 2, 3), "MatrixIndex")
        with self.assertRaises(TypeError):
            check_type(1, "MatrixIndex")
        with self.assertRaises(TypeError):
            check_type((1,), "MatrixIndex")

    def test_non_negative_integer(self):
        assert check_type(2, "NonNegativeInteger")
        assert check_type(0, "NonNegativeInteger")

        with self.assertRaises(MusurgiaTypeError):
            check_type(-2, "NonNegativeInteger")

        with self.assertRaises(MusurgiaTypeError):
            check_type(1.2, "NonNegativeInteger")

    def test_permutation_order(self):
        assert check_type((2, 3, 1), "PermutationOrder")

        with self.assertRaises(MusurgiaTypeError):
            assert check_type([2, 2], "PermutationOrder")
        with self.assertRaises(MusurgiaTypeError):
            assert check_type((2, 2), "PermutationOrder")
        with self.assertRaises(MusurgiaTypeError):
            assert check_type((1, 3, 4), "PermutationOrder")
        with self.assertRaises(MusurgiaTypeError):
            assert check_type((2, 3, 4), "PermutationOrder")
        with self.assertRaises(MusurgiaTypeError):
            assert check_type((1, 3, 4, 2, 4), "PermutationOrder")

    def test_positive_integer(self):
        assert check_type(2, "PositiveInteger")

        with self.assertRaises(MusurgiaTypeError):
            check_type(-2, "PositiveInteger")

        with self.assertRaises(MusurgiaTypeError):
            check_type(0, "PositiveInteger")

        with self.assertRaises(MusurgiaTypeError):
            check_type(-1, "PositiveInteger")

        with self.assertRaises(MusurgiaTypeError):
            check_type(1.2, "PositiveInteger")

    def test_matrix_transpose_type(self):
        assert check_type("regular", "MatrixTransposeMode")
        assert check_type("diagonal", "MatrixTransposeMode")
        with self.assertRaises(MusurgiaTypeError):
            assert check_type("wrong", "MatrixTransposeMode")

    def test_fractal_tree_reduce_children_mode(self):
        assert check_type("sieve", "FractalTreeReduceChildrenMode")
        with self.assertRaises(MusurgiaTypeError):
            assert check_type("wrong", "FractalTreeReduceChildrenMode")

    def test_convertible_to_fraction_type(self):
        assert check_type(1, "ConvertibleToFraction")
        with self.assertRaises(MusurgiaTypeError):
            assert check_type("wrong", "ConvertibleToFraction")

    def test_convert_to_fraction(self):
        assert convert_to_fraction(1) == Fraction(1, 1)
        assert convert_to_fraction(1.5) == Fraction(3, 2)
        assert convert_to_fraction(Fraction(3, 4)) == Fraction(3, 4)

    def test_convertible_to_float(self):
        assert check_type(1, "ConvertibleToFloat")
        assert check_type(1 / 10, "ConvertibleToFloat")
        assert check_type("1.6", "ConvertibleToFloat")
        assert check_type(Fraction(1 / 3), "ConvertibleToFloat")
        with self.assertRaises(MusurgiaTypeError):
            assert check_type("wrong", "ConvertibleToFloat")

    def test_check_label_placement_type(self):
        assert check_type("above", "LabelPlacement")
        with self.assertRaises(MusurgiaTypeError):
            assert check_type("wrong", "LabelPlacement")

    def test_check_pdf_unit_type(self):
        assert check_type("pt", "PdfUnitType")
        with self.assertRaises(MusurgiaTypeError):
            assert check_type("wrong", "PdfUnitType")

    def test_check_position_type(self):
        assert check_type("x", "PositionType")
        with self.assertRaises(MusurgiaTypeError):
            assert check_type("wrong", "PositionType")


class TestCheckValues(TestCase):
    def test_matrix_index(self):
        # wrong index
        with self.assertRaises(ValueError):
            check_matrix_index_values(
                index=(2, 3), number_of_rows=3, number_of_columns=2
            )
        with self.assertRaises(ValueError):
            check_matrix_index_values(
                index=(4, 2), number_of_rows=3, number_of_columns=2
            )

        assert check_matrix_index_values(
            index=(2, 3), number_of_rows=3, number_of_columns=4
        )

    def test_permutation_order(self):
        assert check_permutation_order_values((4, 3, 1, 2), size=4)
        with self.assertRaises(ValueError):
            assert check_permutation_order_values((4, 3, 1, 2), size=5)
        with self.assertRaises(ValueError):
            assert check_permutation_order_values((4, 3, 1, 2), size=3)


class TestLiteralCheckGenerator(TestCase):
    def test_literal_check(self):
        checker = LiteralCheckGenerator(
            HorizontalVertical, "HorizontalVertical"
        ).generate_checker()
        assert checker("h")
        assert checker("v")
        with self.assertRaises(TypeError):
            checker("d")


class TestBuiltInTypeChecker(TestCase):
    def test_builtin_checker(self):
        check_type(v=True, t=bool)
        with self.assertRaises(MusurgiaTypeError):
            check_type(v=True, t=int)



from math import floor, ceil

from fractions import Fraction
from typing import Literal

from musurgia.musurgia_types import ConvertibleToFraction, convert_to_fraction


def _find_nearest_quantized_value(
    quantized_values: list[Fraction], values: list[Fraction]
) -> list[tuple[Fraction, Fraction]]:
    output = []
    for value in values:
        nearest_quantized = min(
            enumerate(quantized_values), key=lambda x: abs(x[1] - value)
        )[1]
        delta = nearest_quantized - value
        output.append((nearest_quantized, delta))
    return output


def _find_quantized_locations(
    positions: list[Fraction], grid_size: ConvertibleToFraction
) -> list[Fraction]:
    grid_size = convert_to_fraction(grid_size)

    def get_quantized(val: Fraction, key: Literal["min", "max"]) -> Fraction:
        factor = Fraction(1, grid_size)
        if key == "min":
            output = Fraction(ceil(val * factor), factor)
        elif key == "max":
            output = Fraction(floor(val * factor), factor)
        else:
            raise ValueError()
        return output

    min_location = Fraction(get_quantized(min(positions), "min"))
    output = [min_location]
    max_location = Fraction(get_quantized(max(positions), "max"))

    while output[-1] < max_location:
        output.append(output[-1] + Fraction(grid_size))

    return output


def get_quantized_positions(
    positions: list[Fraction], grid_size: ConvertibleToFraction
) -> list[Fraction]:
    def _get_quantized_locations() -> list[Fraction]:
        return _find_quantized_locations(positions, grid_size)

    quantized_positions = [
        f[0]
        for f in _find_nearest_quantized_value(_get_quantized_locations(), positions)
    ]

    return quantized_positions


def get_quantized_values(
    values: list[Fraction], grid_size: ConvertibleToFraction
) -> list[Fraction]:
    # print("vals", vals)
    def _get_positions() -> list[Fraction]:
        positions = [Fraction(0)]
        for val in values:
            positions.append(positions[-1] + val)
        return positions

    positions = _get_positions()
    quantized_positions = get_quantized_positions(positions, grid_size)
    quantized_vals = []
    for index in range(len(values)):
        quantized_val = Fraction(
            quantized_positions[index + 1] - quantized_positions[index]
        ).limit_denominator(1000)
        quantized_vals.append(quantized_val)

    return quantized_vals


def find_best_quantized_values(
    values: list[Fraction],
    list_of_grids: list[ConvertibleToFraction],
    check_sum: bool = True,
) -> list[Fraction]:
    if check_sum:
        new_value_grids = []
        for grid in list_of_grids:
            if sum(values) % grid == 0:
                new_value_grids.append(grid)
        if not new_value_grids:
            raise AttributeError("all duration_units failed check_sum")
    else:
        new_value_grids = list_of_grids
    output = values
    old_delta = None
    for grid in new_value_grids:
        quantized_values = get_quantized_values(values, grid)
        new_delta = sum(
            [
                abs(quantized - original)
                for quantized, original in zip(quantized_values, values)
            ]
        )
        if old_delta is None or new_delta < old_delta:
            old_delta = new_delta
            output = quantized_values
    return output



from pathlib import Path
from unittest import TestCase

from musurgia.pdf.line import VerticalLineSegment
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.pdf_tools import draw_ruler
from musurgia.tests.utils_for_tests import PdfTestCase

path = Path(__file__)


class TestVerticalLineSegment(TestCase):
    def setUp(self) -> None:
        self.vls = VerticalLineSegment(length=10)

    def test_get_relative_x2(self):
        assert self.vls.get_relative_x2() == self.vls.get_width()

    def test_start_mark_line_relative_x(self):
        assert self.vls.start_mark_line.relative_x == 0

    def test_start_mark_line_relative_y(self):
        assert self.vls.relative_y == self.vls.start_mark_line.relative_y

    def test_end_mark_line_relative_y(self):
        assert (
            self.vls.end_mark_line.relative_y == self.vls.relative_y + self.vls.length
        )

    def test_end_mark_line_relative_x(self):
        assert self.vls.end_mark_line.relative_x == 0

    def test_start_mark_line_top_margin(self):
        assert self.vls.start_mark_line.top_margin == 0

    def test_get_width(self):
        assert self.vls.get_width() == 3

    def test_set_straight_line_y_x(self):
        self.vls.set_straight_line_relative_x(0)
        assert self.vls.relative_x == -1.5
        with self.assertRaises(NotImplementedError):
            self.vls.set_straight_line_relative_y(0)


class TestVerticalLineSegmentDraw(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.vls = VerticalLineSegment(length=10)

    def test_draw(self):
        self.vls.set_straight_line_relative_x(0)
        with self.file_path(parent_path=path, name="draw", extension="pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, mode="v")
            draw_ruler(self.pdf, mode="h")
            self.pdf.translate(10, 10)
            self.vls.end_mark_line.show = True
            self.vls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)



from fractions import Fraction
from typing import Any, Optional, Union, TypeVar, cast

from musicscore import QuarterDuration, Metronome
from musurgia.musurgia_types import ConvertibleToFraction, check_type, ClockMode
from musurgia.timing.clock import Clock

T = TypeVar("T", bound="ReadonlyDuration")

ConvertibleToDuration = Union["ReadonlyDuration", ConvertibleToFraction]


def _convert_other_to_fraction(other: ConvertibleToDuration) -> Fraction:
    if isinstance(other, ReadonlyDuration):
        return other.calculate_in_seconds()
    check_type(other, "ConvertibleToFraction", function_name="_convert_other")

    return Fraction(other)


def _convert_other_to_duration(other: ConvertibleToDuration) -> "ReadonlyDuration":
    if isinstance(other, ReadonlyDuration):
        return other
    return ReadonlyDuration(other)


class ReadonlyDuration:
    def __init__(
        self,
        seconds: ConvertibleToFraction = 0,
        minutes: ConvertibleToFraction = 0,
        hours: ConvertibleToFraction = 0,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._seconds: Fraction = Fraction(0)
        self._minutes: Fraction = Fraction(0)
        self._hours: Fraction = Fraction(0)
        self._clock: Clock
        self._set_clock(hours=hours, minutes=minutes, seconds=seconds)

    def _set_clock(
        self,
        hours: ConvertibleToFraction,
        minutes: ConvertibleToFraction,
        seconds: ConvertibleToFraction,
    ) -> None:
        self._add_seconds(seconds)
        self._add_minutes(minutes)
        self._add_hours(hours)
        self._clock = Clock.convert_seconds_to_clock(
            Clock.convert_clock_to_seconds(hours, minutes, seconds)
        )
        self._hours, self._minutes, self._seconds = [
            Fraction(time) for time in self._clock.get_values()
        ]

    def _add_seconds(self, val: ConvertibleToFraction) -> None:
        check_type(
            val,
            "ConvertibleToFraction",
            class_name=self.__class__.__name__,
            property_name="seconds",
        )
        self._seconds += Fraction(val)

    def _add_hours(self, val: ConvertibleToFraction) -> None:
        check_type(
            val,
            "ConvertibleToFraction",
            class_name=self.__class__.__name__,
            property_name="hours",
        )
        self._hours += Fraction(val)

    def _add_minutes(self, val: ConvertibleToFraction) -> None:
        check_type(
            val,
            "ConvertibleToFraction",
            class_name=self.__class__.__name__,
            property_name="minutes",
        )
        self._minutes += Fraction(val)

    @property
    def clock(self) -> Clock:
        return self._clock

    @clock.setter
    def clock(self, value: Any) -> None:
        raise AttributeError("ReadonlyDuration cannot set clock.")

    @property
    def minutes(self) -> Fraction:
        return self._minutes

    @minutes.setter
    def minutes(self, value: Any) -> None:
        raise AttributeError("ReadonlyDuration cannot set minutes.")

    @property
    def seconds(self) -> Fraction:
        return self._seconds

    @seconds.setter
    def seconds(self, value: Any) -> None:
        raise AttributeError("ReadonlyDuration cannot set seconds.")

    @property
    def hours(self) -> Fraction:
        return self._hours

    @hours.setter
    def hours(self, value: Any) -> None:
        raise AttributeError("ReadonlyDuration cannot set hours.")

    def calculate_in_seconds(self) -> Fraction:
        return self.seconds + (60 * self.minutes) + (3600 * self.hours)

    def calculate_in_minutes(self) -> Fraction:
        return Fraction(self.seconds, 60) + self.minutes + (60 * self.hours)

    def calculate_in_hours(self) -> Fraction:
        return Fraction(self.seconds, 3600) + Fraction(self.minutes, 60) + self.hours

    def get_clock_as_string(
        self, mode: ClockMode = "hms", round_: Optional[int] = None
    ) -> str:
        return self.clock.get_as_string(mode, round_)

    def __abs__(self: T) -> T:
        return self.__class__(self.calculate_in_seconds().__abs__())

    def __add__(self: T, other: ConvertibleToDuration) -> T:
        return self.__class__(
            self.calculate_in_seconds().__add__(
                _convert_other_to_duration(other).calculate_in_seconds()
            )
        )

    def __ceil__(self: T) -> T:
        return self.__class__(self.calculate_in_seconds().__ceil__())

    def __eq__(self, other: Any) -> bool:
        if other is None:
            return False
        return self.calculate_in_seconds().__eq__(_convert_other_to_fraction(other))

    def __floor__(self: T) -> T:
        return self.__class__(self.calculate_in_seconds().__floor__())

    def __floordiv__(self: T, other: ConvertibleToDuration) -> T:
        return self.__class__(
            self.calculate_in_seconds().__floordiv__(_convert_other_to_fraction(other))
        )

    def __gt__(self, other: ConvertibleToDuration) -> bool:
        return self.calculate_in_seconds().__gt__(_convert_other_to_fraction(other))

    def __ge__(self, other: ConvertibleToDuration) -> bool:
        return self.calculate_in_seconds().__ge__(_convert_other_to_fraction(other))

    def __le__(self, other: ConvertibleToDuration) -> bool:
        return self.calculate_in_seconds().__le__(_convert_other_to_fraction(other))

    def __lt__(self, other: ConvertibleToDuration) -> bool:
        return self.calculate_in_seconds().__lt__(_convert_other_to_fraction(other))

    def __mod__(self: T, other: ConvertibleToDuration) -> T:
        return self.__class__(
            self.calculate_in_seconds().__mod__(_convert_other_to_fraction(other))
        )

    def __mul__(self: T, other: ConvertibleToDuration) -> T:
        return self.__class__(
            self.calculate_in_seconds().__mul__(_convert_other_to_fraction(other))
        )

    def __neg__(self: T) -> T:
        return self.__class__(self.calculate_in_seconds().__neg__())

    def __pos__(self: T) -> T:
        return self.__class__(self.calculate_in_seconds().__pos__())

    def __pow__(self: T, power: Union[int, float]) -> T:
        return self.__class__(pow(self.calculate_in_seconds(), power))

    def __radd__(self: T, other: ConvertibleToDuration) -> T:
        return self.__class__(
            self.calculate_in_seconds().__radd__(_convert_other_to_fraction(other))
        )

    def __rfloordiv__(self: T, other: ConvertibleToDuration) -> T:
        return self.__class__(
            self.calculate_in_seconds().__rfloordiv__(_convert_other_to_fraction(other))
        )

    def __rmod__(self: T, other: ConvertibleToDuration) -> T:
        return self.__class__(
            self.calculate_in_seconds().__rmod__(_convert_other_to_fraction(other))
        )

    def __rmul__(self: T, other: ConvertibleToDuration) -> T:
        return self.__class__(
            self.calculate_in_seconds().__rmul__(_convert_other_to_fraction(other))
        )

    def __round__(self: T, n: Optional[int] = None) -> T:
        return self.__class__(float(self.calculate_in_seconds()).__round__(n))

    # def __rpow__(self, other):
    #     return self.__class__(self.calculate_in_seconds().__rpow__(_convert_other(other)))

    def __rtruediv__(self: T, other: ConvertibleToDuration) -> T:
        return self.__class__(
            self.calculate_in_seconds().__rtruediv__(_convert_other_to_fraction(other))
        )

    def __truediv__(self: T, other: ConvertibleToDuration) -> T:
        return self.__class__(
            self.calculate_in_seconds().__truediv__(_convert_other_to_fraction(other))
        )

    def __trunc__(self) -> int:
        return self.calculate_in_seconds().__trunc__()

    def __str__(self) -> str:
        return f"Duration: {self.clock.get_as_string()}"


class Duration(ReadonlyDuration):
    @ReadonlyDuration.clock.setter  # type: ignore[attr-defined, misc]
    def clock(self, val: Clock) -> None:
        check_type(
            val, Clock, class_name=self.__class__.__name__, property_name="clock"
        )
        self._clock = val
        self._hours, self._minutes, self._seconds = [
            Fraction(time) for time in self.clock.get_values()
        ]

    def add_seconds(self, val: ConvertibleToFraction) -> None:
        super()._add_seconds(val)

    def add_hours(self, val: ConvertibleToFraction) -> None:
        super()._add_hours(val)

    def add_minutes(self, val: ConvertibleToFraction) -> None:
        super()._add_minutes(val)

    @ReadonlyDuration.seconds.setter  # type: ignore[attr-defined, misc]
    def seconds(self, val: ConvertibleToFraction) -> None:
        check_type(
            val,
            "ConvertibleToFraction",
            class_name=self.__class__.__name__,
            property_name="seconds",
        )
        self._set_clock(hours=self._hours, minutes=self._minutes, seconds=val)

    @ReadonlyDuration.minutes.setter  # type: ignore[attr-defined, misc]
    def minutes(self, val: ConvertibleToFraction) -> None:
        check_type(
            val,
            "ConvertibleToFraction",
            class_name=self.__class__.__name__,
            property_name="minutes",
        )
        self._set_clock(self._hours, val, self._seconds)

    @ReadonlyDuration.hours.setter  # type: ignore[attr-defined, misc]
    def hours(self, val: ConvertibleToFraction) -> None:
        check_type(
            val,
            "ConvertibleToFraction",
            class_name=self.__class__.__name__,
            property_name="hours",
        )
        self._set_clock(val, self._minutes, self._seconds)


def convert_duration_to_quarter_duration_value(
    metronome: Union[Metronome, int], duration: Union[Duration, float, int, Fraction]
) -> Fraction:
    if isinstance(duration, Duration):
        seconds = duration.calculate_in_seconds()
    else:
        seconds = Fraction(duration)

    if isinstance(metronome, int):
        metronome = Metronome(metronome)

    return (
        seconds * (Fraction(metronome.per_minute) / 60) * Fraction(metronome.beat_unit)
    )


def convert_quarter_duration_to_duration_value(
    metronome: Union[Metronome, int],
    quarter_duration: Union[QuarterDuration, float, int, Fraction],
) -> Fraction:
    if not isinstance(quarter_duration, QuarterDuration):
        quarter_duration = QuarterDuration(quarter_duration)

    if isinstance(metronome, int):
        metronome = Metronome(metronome)

    return cast(Fraction, quarter_duration.value) / (
        Fraction(metronome.per_minute) / 60 * Fraction(metronome.beat_unit)
    )



from unittest import TestCase

from musurgia.trees.fractaltimelinetree import PermutationIndexCalculater


class TestPermutationMatrixIndexCalculator(TestCase):
    def test_calculate_permutation_matrix_index(self):
        """
        pos_indices = {
            'root': (1, 1),
            '1.1': (3, 1), '1.2': (3, 2), '1.3': (3, 3),

            '2.1.1': (1, 1), '2.1.2': (1, 2), '2.1.3': (1, 3),
            '2.2.1': (2, 1), '2.2.2': (2, 2), '2.2.3': (2, 3),
            '2.3.1': (3, 1), '2.3.2': (3, 2), '2.3.3': (3, 3),

            '3.1.1.1': (2, 1), '3.1.1.2': (2, 2), '3.1.1.3': (2, 3),
            '3.1.2.1': (3, 1), '3.1.2.2': (3, 2), '3.1.2.3': (3, 3),
            '3.1.3.1': (1, 1), '3.1.3.2': (1, 2), '3.1.3.3': (1, 3),

            '3.2.1.1': (3, 1), '3.2.1.2': (3, 2), '3.2.1.3': (3, 3),
            '3.2.2.1': (1, 1), '3.2.2.2': (1, 2), '3.2.2.3': (1, 3),
            '3.2.3.1': (2, 1), '3.2.3.2': (2, 2), '3.2.3.3': (2, 3),

            '3.3.1.1': (1, 1), '3.3.1.2': (1, 2), '3.3.1.3': (1, 3),
            '3.3.2.1': (2, 1), '3.3.2.2': (2, 2), '3.3.2.3': (2, 3),
            '3.3.3.1': (3, 1), '3.3.3.2': (3, 2), '3.3.3.3': (3, 3),
        }
        """
        pic = PermutationIndexCalculater(3)

        pic.parent_index = (1, 1)
        assert pic.get_index(column_number=1) == (2, 1)
        assert pic.get_index(column_number=2) == (2, 2)
        assert pic.get_index(column_number=3) == (2, 3)

        pic.parent_index = (1, 2)
        assert pic.get_index(column_number=1) == (3, 1)
        assert pic.get_index(column_number=2) == (3, 2)
        assert pic.get_index(column_number=3) == (3, 3)

        pic.parent_index = (1, 3)
        assert pic.get_index(column_number=1) == (1, 1)
        assert pic.get_index(column_number=2) == (1, 2)
        assert pic.get_index(column_number=3) == (1, 3)

        pic.parent_index = (2, 1)
        assert pic.get_index(column_number=1) == (3, 1)
        assert pic.get_index(column_number=2) == (3, 2)
        assert pic.get_index(column_number=3) == (3, 3)
        pic.parent_index = (2, 2)
        assert pic.get_index(column_number=1) == (1, 1)
        assert pic.get_index(column_number=2) == (1, 2)
        assert pic.get_index(column_number=3) == (1, 3)

        pic.parent_index = (2, 3)
        assert pic.get_index(column_number=1) == (2, 1)
        assert pic.get_index(column_number=2) == (2, 2)
        assert pic.get_index(column_number=3) == (2, 3)

        pic.parent_index = (3, 1)
        assert pic.get_index(column_number=1) == (1, 1)
        assert pic.get_index(column_number=2) == (1, 2)
        assert pic.get_index(column_number=3) == (1, 3)

        pic.parent_index = (3, 2)
        assert pic.get_index(column_number=1) == (2, 1)
        assert pic.get_index(column_number=2) == (2, 2)
        assert pic.get_index(column_number=3) == (2, 3)

        pic.parent_index = (3, 3)
        assert pic.get_index(column_number=1) == (3, 1)
        assert pic.get_index(column_number=2) == (3, 2)
        assert pic.get_index(column_number=3) == (3, 3)



from typing import Any

from musurgia.musurgia_types import PermutationOrder, check_type, MatrixData
from musurgia.permutation.permutation import get_self_permutation_3d


class LimitedPermutationOrders:
    """
    LimitedPermutationOrders is inspired from Gérard Grisey's permutation technique.
    """

    def __init__(
        self, main_permutation_order: PermutationOrder, *args: Any, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self._main_permutation_order: tuple[int, ...]
        self._permutation_orders: MatrixData
        self.main_permutation_order = main_permutation_order

    @property
    def main_permutation_order(self) -> PermutationOrder:
        return self._main_permutation_order

    @main_permutation_order.setter
    def main_permutation_order(self, val: PermutationOrder) -> None:
        check_type(
            val,
            "PermutationOrder",
            class_name=self.__class__.__name__,
            property_name="main_permutation_order",
        )
        self._main_permutation_order = val
        self._permutation_orders = get_self_permutation_3d(self.main_permutation_order)

    def get_permutation_orders(self) -> MatrixData:
        return self._permutation_orders



from itertools import cycle
from pathlib import Path
from unittest import TestCase

from musicscore.layout import StaffLayout
from musicscore.midi import Midi
from musurgia.magicrandom import MagicRandom
from musurgia.tests.utils_for_tests import (
    XMLTestCase,
    create_test_fractal_relative_musical_tree,
    test_fractal_structur_list,
)
from musurgia.trees.musicaltree import (
    FractalDirectionIterator,
    FractalMusicalTree,
    MidiMusicalTree,
    RelativeMusicTree,
)
from musurgia.trees.timelinetree import TimelineDuration

path = Path(__file__)


class TestSimpleMusicalTree(XMLTestCase):
    def setUp(self):
        self.mt = MidiMusicalTree.create_tree_from_list(
            test_fractal_structur_list, "duration"
        )
        self.mt.get_chord_factory().show_metronome = True

    def test_simple_music_tree_root_chord(self):
        chord = self.mt.get_chord_factory().create_chord()
        self.assertEqual(
            chord.quarter_duration, self.mt.get_duration().get_quarter_duration()
        )
        self.assertEqual(chord.metronome, self.mt.get_duration().get_metronome())

    def test_simple_music_tree_to_score(self):
        score = self.mt.export_score()
        score.get_quantized = True
        with self.file_path(path, "simple") as xml_path:
            score.export_xml(xml_path)


class TestRandomMusicalTree(XMLTestCase):
    def setUp(self):
        self.mt = MidiMusicalTree.create_tree_from_list(
            test_fractal_structur_list, "duration"
        )
        self.mt.get_chord_factory().show_metronome = True

    def set_random_midis(self, musical_tree, root_midi_range, periodicitiy, seed):
        min_midi, max_midi = root_midi_range
        random_ = MagicRandom(
            pool=list(range(min_midi, max_midi + 1)),
            periodicity=periodicitiy,
            seed=seed,
        )
        for node in musical_tree.traverse():
            node.get_chord_factory().midis = [Midi(next(random_))]

    def test_random_midis(self):
        self.set_random_midis(
            musical_tree=self.mt, root_midi_range=(60, 84), seed=10, periodicitiy=7
        )
        score = self.mt.export_score()
        score.get_quantized = True
        with self.file_path(path, "random") as xml_path:
            score.export_xml(xml_path)


class TestRelativeMidiMusicalTree(XMLTestCase):
    def setUp(self):
        self.mt = RelativeMusicTree.create_tree_from_list(
            test_fractal_structur_list, "duration"
        )
        self.mt.midi_value_range = (60, 84)
        self.mt.get_chord_factory().show_metronome = True

    def test_relative_midis(self):
        self.mt.set_relative_midis()
        score = self.mt.export_score()
        score.get_quantized = True
        with self.file_path(path, "relative") as xml_path:
            score.export_xml(xml_path)


class TestFractalDirectionIterator(TestCase):
    def test_fractal_direction_iterator(self):
        ft = FractalMusicalTree(
            duration=TimelineDuration(10),
            proportions=(1, 2, 3, 4),
            main_permutation_order=(3, 1, 4, 2),
            permutation_index=(1, 1),
        )
        ft.add_layer()
        fdi = FractalDirectionIterator(main_direction_cell=[1, -1], fractal_node=ft)
        self.assertEqual(fdi.get_main_directions(), [1, -1, 1, -1])
        self.assertEqual(fdi.get_directions(), [1, 1, -1, -1])


class TestRelativeFractalMusicalTree(XMLTestCase):
    def setUp(self):
        self.ft = create_test_fractal_relative_musical_tree()
        self.ft.get_chord_factory().show_metronome = True

    def test_default_direction_iterator(self):
        # self.ft.set_relative_midis()
        for node in self.ft.traverse():
            self.assertEqual(
                node.direction_iterator.get_main_directions(), [1, -1, 1, -1]
            )
        expected = """└── [1, 1, -1, -1]
    ├── [-1, -1, 1, 1]
    │   ├── []
    │   ├── [-1, -1, 1, 1]
    │   │   ├── []
    │   │   ├── []
    │   │   ├── []
    │   │   └── []
    │   ├── []
    │   └── [1, 1, -1, -1]
    │       ├── []
    │       ├── []
    │       ├── []
    │       └── []
    ├── []
    ├── [1, -1, 1, -1]
    │   ├── []
    │   ├── []
    │   ├── [-1, -1, 1, 1]
    │   │   ├── []
    │   │   ├── []
    │   │   ├── []
    │   │   └── []
    │   └── [1, -1, 1, -1]
    │       ├── []
    │       ├── []
    │       ├── []
    │       └── []
    └── [-1, 1, -1, 1]
        ├── [-1, -1, 1, 1]
        │   ├── []
        │   ├── []
        │   ├── []
        │   └── []
        ├── [1, 1, -1, -1]
        │   ├── []
        │   ├── []
        │   ├── []
        │   └── []
        ├── []
        └── []
"""
        self.assertEqual(
            self.ft.get_tree_representation(
                key=lambda node: node.direction_iterator.get_directions()
            ),
            expected,
        )

    def test_relative_fractal_musical_tree(self):
        score = self.ft.export_score()
        score.staff_layout = StaffLayout()
        score.staff_layout.staff_distance = 100
        score.get_quantized = True
        with self.file_path(path, "fractal_relative") as xml_path:
            score.export_xml(xml_path)

    def test_relative_fractat_musical_ziczac_tree(self):
        for node in self.ft.traverse():
            node.direction_iterator = cycle([-1, 1])
        self.ft.set_relative_midis()
        score = self.ft.export_score()
        score.staff_layout = StaffLayout()
        score.staff_layout.staff_distance = 100

        score.get_quantized = True
        with self.file_path(path, "fractal_relative_ziczac") as xml_path:
            score.export_xml(xml_path)

    def test_set_relatve_midis_again(self):
        expected = """└── 60
    ├── 68.0
    │   ├── 80.0
    │   ├── 76.0
    │   │   ├── 68.0
    │   │   ├── 71.0
    │   │   ├── 76.0
    │   │   └── 75.0
    │   ├── 68.0
    │   └── 70.0
    │       ├── 72.0
    │       ├── 75.0
    │       ├── 76.0
    │       └── 72.0
    ├── 80.0
    ├── 84.0
    │   ├── 76.0
    │   ├── 72.0
    │   ├── 80.0
    │   │   ├── 68.0
    │   │   ├── 72.0
    │   │   ├── 80.0
    │   │   └── 78.0
    │   └── 68.0
    │       ├── 76.0
    │       ├── 80.0
    │       ├── 72.0
    │       └── 84.0
    └── 68.0
        ├── 60.0
        │   ├── 68.0
        │   ├── 65.0
        │   ├── 60.0
        │   └── 61.0
        ├── 68.0
        │   ├── 66.0
        │   ├── 63.0
        │   ├── 62.0
        │   └── 66.0
        ├── 62.0
        └── 66.0
"""
        self.assertEqual(
            self.ft.get_tree_representation(
                key=lambda node: node.get_chord_factory().midis[0].value
            ),
            expected,
        )
        for node in self.ft.traverse():
            node.direction_iterator.reset()
        self.ft.set_relative_midis()
        self.assertEqual(
            self.ft.get_tree_representation(
                key=lambda node: node.get_chord_factory().midis[0].value
            ),
            expected,
        )



from unittest import TestCase

from musicscore.chord import XMLWords

from musurgia.chordfactory.chordfactory import AbstractChordFactory


class SimpleDemoChordFactory(AbstractChordFactory):
    def update_chord_quarter_duration(self):
        self._chord.quarter_duration.value = 4

    def update_chord_midis(self):
        self._chord.midis = 72

    def update_chord_words(self):
        self._chord.add_x(XMLWords("something"))


class ChordFactoryTestCase(TestCase):
    def test_simple_demo_chord_factory(self):
        chf = SimpleDemoChordFactory()
        ch = chf.create_chord()
        self.assertEqual(ch.midis[0].value, 72)
        self.assertEqual(ch.quarter_duration, 4)
        self.assertEqual(ch.get_words()[0].value_, "something")



from pathlib import Path

from musurgia.pdf.line import MarkLine
from musurgia.pdf.pdf import Pdf
from musurgia.tests.utils_for_tests import PdfTestCase, DummyMaster

path = Path(__file__)


class TestMarkLine(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.master = DummyMaster()
        self.ml = MarkLine(
            mode="v", placement="start", name="mark_line_test", master=self.master
        )

    def test_relative_x(self):
        actual = self.ml.relative_x
        expected = self.master.get_slave_position(self.ml, "x")
        self.assertEqual(expected, actual)

    def test_get_relative_x2(self):
        actual = self.ml.get_relative_x2()
        expected = self.ml.relative_x
        self.assertEqual(expected, actual)

    def test_get_relative_y2(self):
        actual = self.ml.get_relative_y2()
        expected = self.ml.relative_y + self.ml.length
        self.assertEqual(expected, actual)

    def test_get_height(self):
        actual = self.ml.get_height()
        expected = self.ml.top_margin + self.ml.length + self.ml.bottom_margin
        self.assertEqual(expected, actual)

    def test_placement(self):
        assert self.ml.placement == "start"
        self.ml.placement = "end"
        assert self.ml.placement == "end"

    def test_draw(self):
        with self.file_path(path, "draw", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            self.ml.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_draw_multiple(self):
        self.ml.master.get_slave_position = lambda slave, position: 0
        self.ml.master.get_slave_margin = lambda slave, margin: 0
        with self.file_path(path, "draw_multiple", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            self.ml.draw(self.pdf)
            self.ml.master.get_slave_margin = (
                lambda slave, margin: 5 if margin in ["l", "left"] else 0
            )
            self.ml.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)



from fractions import Fraction
import os
import unittest
from pathlib import Path

from diff_pdf_visually import pdf_similar  # type: ignore

from musurgia.trees.fractaltimelinetree import FractalTimelineTree
from musurgia.pdf import TextLabel, DrawObjectColumn, StraightLine
from musurgia.pdf.drawobject import MasterDrawObject
from musurgia.trees.musicaltree import FractalMusicalTree, FractalRelativeMusicTree, MidiMusicalTreeChordFactory
from musurgia.trees.timelinetree import TimelineDuration, TimelineTree
from musurgia.trees.valuedtree import ValuedTree

test_fractal_structur_list = [
    Fraction(10, 1),
    [
        Fraction(3, 1),
        [Fraction(3, 5)],
        [
            Fraction(6, 5),
            [Fraction(6, 25)],
            [Fraction(12, 25)],
            [Fraction(3, 25)],
            [Fraction(9, 25)],
        ],
        [Fraction(3, 10)],
        [
            Fraction(9, 10),
            [Fraction(27, 100)],
            [Fraction(9, 100)],
            [Fraction(9, 25)],
            [Fraction(9, 50)],
        ],
    ],
    [Fraction(1, 1)],
    [
        Fraction(4, 1),
        [Fraction(2, 5)],
        [Fraction(4, 5)],
        [
            Fraction(6, 5),
            [Fraction(6, 25)],
            [Fraction(12, 25)],
            [Fraction(3, 25)],
            [Fraction(9, 25)],
        ],
        [
            Fraction(8, 5),
            [Fraction(4, 25)],
            [Fraction(8, 25)],
            [Fraction(12, 25)],
            [Fraction(16, 25)],
        ],
    ],
    [
        Fraction(2, 1),
        [
            Fraction(4, 5),
            [Fraction(4, 25)],
            [Fraction(8, 25)],
            [Fraction(2, 25)],
            [Fraction(6, 25)],
        ],
        [
            Fraction(3, 5),
            [Fraction(9, 50)],
            [Fraction(3, 50)],
            [Fraction(6, 25)],
            [Fraction(3, 25)],
        ],
        [Fraction(2, 5)],
        [Fraction(1, 5)],
    ],
]


def create_test_path(path, test_name):
    return path.parent.joinpath(f"{path.stem}_{test_name}")


class FilePath:
    def __init__(self, unittest, parent_path, name, extension):
        self._unittest = None
        self._parent_path = None
        # self._name = None
        # self._extension = None

        self.unittest = unittest
        self.parent_path = parent_path
        self.name = name
        self.extension = extension
        self.out_path = None

    @property
    def unittest(self):
        return self._unittest

    @unittest.setter
    def unittest(self, val):
        # if not isinstance(val, PdfTestCase):
        #     raise TypeError(f"unittest.value must be of type {type(PdfTestCase)} not{type(val)}")
        self._unittest = val

    @property
    def parent_path(self):
        return self._parent_path

    @parent_path.setter
    def parent_path(self, val):
        if not isinstance(val, Path):
            raise TypeError(
                f"parent_path.value must be of type {type(Path)} not{type(val)}"
            )
        self._parent_path = val

    def __enter__(self):
        self.out_path = create_test_path(
            self.parent_path, self.name + "." + self.extension
        )
        return self.out_path

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unittest.assertCompareFiles(self.out_path)


class PdfTestCase(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _compare_pdfs(self, actual_file_path, expected_file_path, verbosity):
        self.assertTrue(
            pdf_similar(actual_file_path, expected_file_path, verbosity=verbosity)
        )

    def _compare_contents(self, actual_file_path, expected_file_path):
        with open(actual_file_path, "r") as myfile:
            result = myfile.read()

        with open(expected_file_path, "r") as myfile:
            expected = myfile.read()

        self.assertEqual(expected, result)

    def assertCompareFiles(
        self, actual_file_path, expected_file_path=None, verbosity=0
    ):
        file_name, extension = os.path.splitext(actual_file_path)
        if not expected_file_path:
            if not extension:
                raise ValueError("actual_file_path has no file extension")
            expected_file_path = file_name + "_expected" + extension

        if extension == ".pdf":
            self._compare_pdfs(
                actual_file_path, expected_file_path, verbosity=verbosity
            )
        else:
            self._compare_contents(actual_file_path, expected_file_path)

    def file_path(self, parent_path, name, extension):
        tfp = FilePath(self, parent_path, name, extension)
        return tfp


class XMLTestCase(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _compare_contents(self, actual_file_path, expected_file_path):
        with open(actual_file_path, "r") as myfile:
            result = myfile.read()

        with open(expected_file_path, "r") as myfile:
            expected = myfile.read()

        self.assertEqual(expected, result)

    def assertCompareFiles(self, actual_file_path, expected_file_path=None):
        file_name, extension = os.path.splitext(actual_file_path)
        if not expected_file_path:
            if not extension:
                expected_file_path += ".xml"
            expected_file_path = file_name + "_expected" + ".xml"

        self._compare_contents(actual_file_path, expected_file_path)

    def file_path(self, parent_path, name):
        tfp = FilePath(self, parent_path, name, "xml")
        return tfp


def fractal_node_info(node):
    return f"{node.get_fractal_order()}: {node.get_permutation_index()}: {round(float(node.get_value()), 2)}"


def fractal_node_info_with_permutation_order(node):
    return f"{node.get_fractal_order()}: {node.get_permutation_index()}: {node.get_permutation_order()}: {round(float(node.get_value()), 2)}"


class DummyMaster(MasterDrawObject):
    def get_slave_margin(self, slave, margin):
        return 10

    def get_slave_position(self, slave, position):
        return 20

    def draw(self, pdf):
        pass

    def get_relative_x2(self):
        pass

    def get_relative_y2(self):
        pass


def add_control_positions_to_draw_object(draw_object):
    text_1 = TextLabel(draw_object.positions, font_size=8)
    text_2 = TextLabel(
        draw_object.get_end_positions(),
        placement="below",
        font_size=8,
        left_margin=(draw_object.get_relative_x2() - draw_object.relative_x),
    )

    draw_object.add_text_label(text_1)
    draw_object.add_text_label(text_2)


numbers = ["one", "two", "three"]


def add_test_left_labels(drawobject):
    for t in numbers:
        drawobject.add_text_label(f"left label {t}", placement="left", font_size=8)


def add_test_below_labels(drawobject):
    for t in numbers:
        drawobject.add_text_label(f"below label {t}", placement="below", font_size=8)


# def add_test_right_labels(drawobject):
#     for t in numbers:
#         drawobject.add_text_label(f'right label {t}', placement='right', font_size=8)


def add_test_above_labels(drawobject):
    for t in numbers:
        drawobject.add_text_label(f"above label {t}", placement="above", font_size=8)


def add_test_right_above_labels(drawobject):
    for t in numbers:
        drawobject.add_text_label(
            f"right above label {t}", placement="right_above", font_size=8
        )


def add_test_right_below_labels(drawobject):
    for t in numbers:
        drawobject.add_text_label(
            f"right below label {t}", placement="right_below", font_size=8
        )


def add_test_labels(drawobject):
    add_test_above_labels(drawobject)
    add_test_below_labels(drawobject)
    add_test_left_labels(drawobject)
    # add_test_right_labels(drawobject)
    # add_test_right_above_labels(drawobject)
    # add_test_right_below_labels(drawobject)


def create_simple_column(list_of_draw_objects):
    c = DrawObjectColumn(show_borders=True, show_margins=True)
    for do in list_of_draw_objects:
        do.relative_y = 5
        do.margins = (10, 10, 10, 10)
    divider = StraightLine("h", max([do.get_width() for do in list_of_draw_objects]))
    for do in list_of_draw_objects:
        c.add_draw_object(do)
        if do != list_of_draw_objects[-1]:
            c.add_draw_object(divider)

    return c


def create_test_fractal_timline_tree():
    ft = FractalTimelineTree(
        duration=TimelineDuration(10),
        proportions=(1, 2, 3, 4),
        main_permutation_order=(3, 1, 4, 2),
        permutation_index=(1, 1),
    )
    ft.add_layer()

    ft.add_layer(lambda node: node.get_fractal_order() > 1)
    ft.add_layer(lambda node: node.get_fractal_order() > 2)
    return ft


def create_test_fractal_musical_tree():
    ft = FractalMusicalTree(
        duration=TimelineDuration(10),
        proportions=(1, 2, 3, 4),
        main_permutation_order=(3, 1, 4, 2),
        permutation_index=(1, 1),
    )
    ft.add_layer()

    ft.add_layer(lambda node: node.get_fractal_order() > 1)
    ft.add_layer(lambda node: node.get_fractal_order() > 2)
    return ft

class FractalMusicalTreeChordFactory(MidiMusicalTreeChordFactory):
    def update_chord_fractal_order(self):
        self._chord.add_words(str(self.get_musical_tree_node().get_fractal_order()), placement="below", relative_y=-10)

class MyFractalRelativeMusicTree(FractalRelativeMusicTree):
    TreeChordFactoryClass = FractalMusicalTreeChordFactory

def create_test_fractal_relative_musical_tree():
    ft = MyFractalRelativeMusicTree(
        duration=TimelineDuration(10),
        proportions=(1, 2, 3, 4),
        main_permutation_order=(3, 1, 4, 2),
        permutation_index=(1, 1),
        midi_value_range=(60, 84),
    )
    ft.add_layer()

    ft.add_layer(lambda node: node.get_fractal_order() > 1)
    ft.add_layer(lambda node: node.get_fractal_order() > 2)
    ft.set_relative_midis()
    return ft


def create_test_timeline_tree():
    tlt = TimelineTree(duration=TimelineDuration(60))
    for d in [20, 10, 30]:
        tlt.add_child(TimelineTree(TimelineDuration(d)))
    for list_of_d, child in zip([[10, 2, 3, 5], [], [5, 20, 3, 2]], tlt.get_children()):
        for d in list_of_d:
            child.add_child(TimelineTree(TimelineDuration(d)))
    return tlt


class DemoValuedTree(ValuedTree):
    def __init__(self, value, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._value = value

    def _check_child_to_be_added(self, child):
        if not isinstance(child, ValuedTree):
            raise TypeError
        else:
            return True

    def _set_value(self, value):
        self._value = Fraction(value)

    def get_value(self):
        return Fraction(self._value)


def create_test_valued_tree() -> DemoValuedTree:
    vt = DemoValuedTree.create_tree_from_list(test_fractal_structur_list, "value")
    # print(vt.get_tree_representation(key=lambda node: node.get_value()))

    """└── 10
    ├── 3
    │   ├── 3/5
    │   ├── 6/5
    │   │   ├── 6/25
    │   │   ├── 12/25
    │   │   ├── 3/25
    │   │   └── 9/25
    │   ├── 3/10
    │   └── 9/10
    │       ├── 27/100
    │       ├── 9/100
    │       ├── 9/25
    │       └── 9/50
    ├── 1
    ├── 4
    │   ├── 2/5
    │   ├── 4/5
    │   ├── 6/5
    │   │   ├── 6/25
    │   │   ├── 12/25
    │   │   ├── 3/25
    │   │   └── 9/25
    │   └── 8/5
    │       ├── 4/25
    │       ├── 8/25
    │       ├── 12/25
    │       └── 16/25
    └── 2
        ├── 4/5
        │   ├── 4/25
        │   ├── 8/25
        │   ├── 2/25
        │   └── 6/25
        ├── 3/5
        │   ├── 9/50
        │   ├── 3/50
        │   ├── 6/25
        │   └── 3/25
        ├── 2/5
        └── 1/5
    """
    return vt


def add_node_infos_to_graphic(vt, gt):
    for gn, vn in zip(gt.traverse(), vt.traverse()):
        value_label = gn.get_segment().start_mark_line.add_text_label(
            round(float(vn.get_value()), 1), font_size=8, bottom_margin=1
        )

        position_in_tree_label = gn.get_segment().start_mark_line.add_text_label(
            f"{vn.get_position_in_tree()}", font_size=8, placement="below", top_margin=1
        )
        if gn.get_distance() == 3:
            if vn.up.up.get_position_in_tree() == "4":
                if gn.up.get_children().index(gn) % 4 == 1:
                    position_in_tree_label.top_margin = 3
                    value_label.bottom_margin = 4
                elif gn.up.get_children().index(gn) % 4 == 2:
                    position_in_tree_label.top_margin = 5
                    value_label.bottom_margin = 7
                elif gn.up.get_children().index(gn) % 4 == 3:
                    position_in_tree_label.top_margin = 7
                    value_label.bottom_margin = 10
            else:
                if gn.up.get_children().index(gn) % 2 == 1:
                    position_in_tree_label.top_margin = 3
                    value_label.bottom_margin = 4



from unittest import TestCase

from musurgia.trees.fractaltimelinetree import FractalTimelineTree
from musurgia.trees.timelinetree import TimelineDuration


class TestFtMergeChildren(TestCase):
    def test_one_layer(self):
        ft = FractalTimelineTree(
            proportions=(1, 2, 3, 4, 5),
            main_permutation_order=(3, 5, 1, 2, 4),
            duration=TimelineDuration(10),
            permutation_index=(1, 1),
        )
        ft.add_layer()
        ft.merge_children(1, 2, 2)
        self.assertEqual(
            [3, 5, 2],
            ft.get_leaves(key=lambda leaf: leaf.get_fractal_order()),
        )
        self.assertEqual(
            [2.0, 4.0, 4.0],
            ft.get_leaves(key=lambda leaf: round(float(leaf.get_value()), 2)),
        )

    def test_two_layers(self):
        ft = FractalTimelineTree(
            proportions=(1, 2, 3, 4, 5),
            main_permutation_order=(3, 5, 1, 2, 4),
            duration=TimelineDuration(20),
            permutation_index=(1, 1),
        )
        ft.add_layer()
        ft.add_layer()
        # print(ft.get_layer(1, key=lambda leaf: round(float(leaf.get_value()), 2)))
        #  [4.0, 6.67, 1.33, 2.67, 5.33]
        # pprint(ft.get_layer(2, key=lambda leaf: round(float(leaf.get_value()), 2)))
        """
        [[0.8, 0.53, 0.27, 1.07, 1.33],
         [1.33, 1.78, 0.44, 2.22, 0.89],
         [0.27, 0.44, 0.09, 0.18, 0.36],
         [0.18, 0.71, 0.53, 0.89, 0.36],
         [0.36, 1.78, 1.07, 0.71, 1.42]]
        """
        ft.merge_children(1, 2, 2)
        self.assertEqual(
            [4.0, 8.0, 8.0],
            ft.get_layer(1, key=lambda leaf: round(float(leaf.get_value()), 2)),
        )
        self.assertEqual(
            [
                0.8,
                0.53,
                0.27,
                1.07,
                1.33,
                1.6,
                2.13,
                0.53,
                2.67,
                1.07,
                0.53,
                2.13,
                1.6,
                2.67,
                1.07,
            ],
            ft.get_layer(2, key=lambda leaf: round(float(leaf.get_value()), 2)),
        )



from fractions import Fraction
from typing import Any, Optional, Union, Literal, Callable, cast, get_args

from musurgia.musurgia_exceptions import MatrixIndexOutOfRangeError

MUSURGIA_TYPES = [
    "MatrixData",
    "MatrixIndex",
    "MatrixTransposeMode",
    "NonNegativeInteger",
    "PermutationOrder",
    "PositiveInteger",
    "ConvertibleToFraction",
    "FractalTreeReduceChildrenMode",
    "MatrixReadingDirection",
    "ConvertibleToFloat",
    "LabelPlacement",
    "HorizontalVertical",
    "PdfUnitType",
    "PositionType",
    "FontFamily",
    "FontWeight",
    "FontStyle",
    "VerticalPosition",
    "HorizontalPosition",
    "MarkLinePlacement",
    "PageOrientation",
    "PageFormat",
    "PageOrientation",
    "MarginType",
    "ClockMode",
]

MusurgiaType = Literal[
    "MatrixData",
    "MatrixIndex",
    "MatrixTransposeMode",
    "NonNegativeInteger",
    "PermutationOrder",
    "PositiveInteger",
    "ConvertibleToFraction",
    "ConvertibleToFloat",
    "FractalTreeReduceChildrenMode",
    "MatrixReadingDirection",
    "LabelPlacement",
    "HorizontalVertical",
    "PdfUnitType",
    "PositionType",
    "FontFamily",
    "FontWeight",
    "FontStyle",
    "VerticalPosition",
    "HorizontalPosition",
    "MarkLinePlacement",
    "PageOrientation",
    "PageFormat",
    "PageOrientation",
    "MarginType",
    "ClockMode",
]


class LiteralCheckGenerator:
    def __init__(self, literal_type: Any, type_name: str):
        self.type_name = type_name
        self.permitted = get_args(literal_type)

    def generate_checker(self) -> Callable[[str], bool]:
        def checker(value: str) -> bool:
            if value not in self.permitted:
                raise TypeError(
                    f"{self.type_name} value must be in {self.permitted}, got {value}"
                )
            return True

        return checker


def create_error_message(
    v: Optional[Any] = None,
    t: Optional[Union[type, str]] = None,
    function_name: Optional[str] = None,
    class_name: Optional[str] = None,
    method_name: Optional[str] = None,
    argument_name: Optional[str] = None,
    property_name: Optional[str] = None,
    class_attribute_name: Optional[str] = None,
    message: Optional[str] = None,
) -> str:
    if not message and not (v or t):
        raise AttributeError("if no message provided v and t must be set")

    if message and (v or t):
        raise AttributeError("if message is provided no v and t can be set")

    if function_name and (property_name or method_name or class_name):
        raise AttributeError(
            "function_name cannot be set with property_name or method_name or class_name"
        )

    if function_name and not argument_name:
        raise AttributeError("After setting function_name argument_name must be set.")

    if class_name and not (property_name or method_name or class_attribute_name):
        raise AttributeError(
            "After setting class_name property_name, method_name or class_attribute_name must be set."
        )

    if class_attribute_name and not class_name:
        raise AttributeError(
            "After setting class_attribute_name class_name must be set."
        )

    if class_attribute_name and (property_name or method_name or argument_name):
        raise AttributeError(
            "class_attribute_name and property_name, method_name or argument_name cannot be set together."
        )

    if method_name and (property_name or class_attribute_name):
        raise AttributeError(
            "method_name and property_name or class_attribute_name cannot be set together."
        )

    if method_name and not message and not (argument_name and class_name):
        raise AttributeError(
            "After setting method_name class_name and argument_name must be set."
        )

    if method_name and message and not class_name:
        raise AttributeError(
            "After setting message and method_name class_name must be set."
        )

    if argument_name and not (function_name or method_name):
        raise AttributeError(
            "After setting argument_name method_name or function_name must be set."
        )

    if argument_name and property_name:
        raise AttributeError("argument_name and property_name cannot be set together.")

    if property_name and not class_name:
        raise AttributeError("After setting property_name class_name must be set.")

    if not message:
        if isinstance(t, str):
            _type = t
        else:
            _type = cast(type, t).__name__
        message = f"Value {v} must be of type {_type} not {v.__class__.__name__}"

    if property_name and class_name:
        msg = f"{class_name}.{property_name}: {message}"

    elif function_name and argument_name:
        msg = f"{function_name}:{argument_name}: {message}"

    elif argument_name and method_name and class_name:
        msg = f"{class_name}.{method_name}:{argument_name}: {message}"

    elif argument_name and method_name and not class_name:
        msg = f"{method_name}:{argument_name}: {message}"

    elif argument_name and not method_name and not class_name:
        msg = f"{argument_name}: {message}"
    else:
        msg = f"{message}"
    return msg


def check_musurgia_type_type(value: str) -> bool:
    if not isinstance(value, str):
        raise TypeError(f"MusurgiaType value must be of type str,  got {type(value)}")
    if value not in MUSURGIA_TYPES:
        raise TypeError(f"MusurgiaType value must be in {MUSURGIA_TYPES}, got {value}")
    return True


class MusurgiaTypeError(TypeError):
    """
    :param t: ``type``
    :param v: ``value``
    :param argument_name: name of the argument which is being checked.
    :param method_name: name of the method which is executing this checking.
    :param obj: ``object`` which the executing method is a part of

    If ``argument_name`` is not set ``method_name`` and ``obj`` have no impact.
    If ``method_name`` is not set ``obj`` has no impact.
    """

    def __init__(
        self,
        v: Optional[Any] = None,
        t: Optional[Union[type, str]] = None,
        function_name: Optional[str] = None,
        class_name: Optional[str] = None,
        method_name: Optional[str] = None,
        argument_name: Optional[str] = None,
        property_name: Optional[str] = None,
        class_attribute_name: Optional[str] = None,
        message: Optional[str] = None,
    ):
        msg = create_error_message(
            v,
            t,
            function_name,
            class_name,
            method_name,
            argument_name,
            property_name,
            class_attribute_name,
            message,
        )
        super().__init__(msg)

    def __setattr__(self, attr: str, value: Any) -> Any:
        raise AttributeError(
            "Trying to set attribute on a frozen instance MusurgiaTypeError"
        )


NonNegativeInteger = int


def check_non_negative_integer_type(value: NonNegativeInteger) -> bool:
    if not isinstance(value, int) or value < 0:
        raise TypeError(
            f"NonNegativeInteger value must be a non-negative integer, got {value}"
        )
    return True


PositiveInteger = int


def check_positive_integer_type(value: PositiveInteger) -> bool:
    if not isinstance(value, int) or value <= 0:
        raise TypeError(
            f"PositiveInteger value must be a positive integer, got {value}"
        )
    return True


ConvertibleToFraction = Union[float, int, Fraction]


def check_convertible_to_fraction_type(value: ConvertibleToFraction) -> bool:
    if (
        not isinstance(value, int)
        and not isinstance(value, Fraction)
        and not isinstance(value, float)
    ):
        raise TypeError
    return True


def convert_to_fraction(value: ConvertibleToFraction) -> Fraction:
    if isinstance(value, Fraction):
        return value
    else:
        return Fraction(value)


ConvertibleToFloat = Union[float, int, Fraction, str]


def check_convertible_to_float_type(value: ConvertibleToFloat) -> bool:
    try:
        float(value)
    except (TypeError, ValueError):
        raise TypeError
    return True


PermutationOrder = tuple[int, ...]


def check_permutation_order_type(value: PermutationOrder) -> bool:
    if (
        not isinstance(value, tuple)
        or len(value) != len(set(value))
        or set(value) != set(range(1, len(value) + 1))
    ):
        raise TypeError(
            f"PermutationOrder value must be a tuple with all integers from 1 to an upper limit corresponding the size of input_list, got {value}"
        )
    return True


def check_permutation_order_values(
    permutation_order: PermutationOrder, size: NonNegativeInteger
) -> bool:
    check_type(
        v=size,
        t="NonNegativeInteger",
        function_name="check_permutation_order_values",
        argument_name="size",
    )
    if len(permutation_order) != size:
        raise ValueError(f"PermutationOrder {permutation_order} must be of size {size}")
    return True


MatrixData = list[list[Any]]


def check_matrix_data_type(matrix_data: MatrixData) -> bool:
    row_size = None
    for i, row in enumerate(matrix_data):
        if not isinstance(row, list):
            raise TypeError(f"row {row} is not a list")
        if i == 0:
            row_size = len(row)
        else:
            if len(row) != row_size:
                raise TypeError(f"row {row} must be of length {row_size}")
    return True


MatrixIndex = tuple[PositiveInteger, PositiveInteger]


def check_matrix_index_type(index: MatrixIndex) -> bool:
    if (
        not isinstance(index, tuple)
        or len(index) != 2
        or not check_positive_integer_type(
            index[0] or not check_positive_integer_type(index[1])
        )
    ):
        raise TypeError(
            f"MatrixIndex: index {index} must be a tuple with two positive integers"
        )
    return True


def check_matrix_index_values(
    index: MatrixIndex,
    number_of_rows: PositiveInteger,
    number_of_columns: PositiveInteger,
) -> bool:
    check_positive_integer_type(number_of_rows)
    check_positive_integer_type(number_of_columns)
    if index[0] > number_of_rows or index[1] > number_of_columns:
        raise MatrixIndexOutOfRangeError(
            f"MatrixIndex: index {index} must be in ranges (1..{number_of_rows}, 1..{number_of_columns}) "
        )

    return True


MatrixTransposeMode = Literal["regular", "diagonal"]
check_matrix_transpose_mode_type = LiteralCheckGenerator(
    MatrixTransposeMode, "MatrixTransposeMode"
).generate_checker()

MatrixReadingDirection = Literal["horizontal", "diagonal", "vertical"]
check_matrix_reading_direction_type = LiteralCheckGenerator(
    MatrixReadingDirection, "MatrixReadingDirection"
).generate_checker()
FractalTreeReduceChildrenMode = Literal["backwards", "forwards", "sieve", "merge"]
check_fractal_tree_reduce_children_mode_type = LiteralCheckGenerator(
    FractalTreeReduceChildrenMode, "FractalTreeReduceChildrenMode"
).generate_checker()

LabelPlacement = Literal["above", "below", "left"]
check_label_placement_type = LiteralCheckGenerator(
    LabelPlacement, "LabelPlacement"
).generate_checker()

HorizontalVertical = Literal["horizontal", "h", "vertical", "v"]
check_horizontal_vertical_type = LiteralCheckGenerator(
    HorizontalVertical, "HorizontalVertical"
).generate_checker()

PdfUnitType = Literal["pt", "mm", "cm", "in"]
check_pdf_unit_type_type = LiteralCheckGenerator(
    PdfUnitType, "PdfUnitType"
).generate_checker()

PositionType = Literal["x", "y"]
check_position_type_type = LiteralCheckGenerator(
    PositionType, "PositionType"
).generate_checker()

FontFamily = Literal["Courier"]
check_font_family_type = LiteralCheckGenerator(
    FontFamily, "FontFamily"
).generate_checker()

FontWeight = Literal["medium", "bold"]
check_font_weight_type = LiteralCheckGenerator(
    FontWeight, "FontWeight"
).generate_checker()

FontStyle = Literal["regular", "italic"]
check_font_style_type = LiteralCheckGenerator(FontStyle, "FontStyle").generate_checker()

VerticalPosition = Literal["top", "bottom"]
check_vertical_position_type = LiteralCheckGenerator(
    VerticalPosition, "VerticalPosition"
).generate_checker()

HorizontalPosition = Literal["left", "center", "right"]
check_horizontal_position_type = LiteralCheckGenerator(
    HorizontalPosition, "HorizontalPosition"
).generate_checker()

MarkLinePlacement = Literal["start", "end"]
check_mark_line_placement_type = LiteralCheckGenerator(
    MarkLinePlacement, "MarkLinePlacement"
).generate_checker()

PageOrientation = Literal["", "portrait", "p", "P", "landscape", "l", "L"]
check_page_orientation_type = LiteralCheckGenerator(
    PageOrientation, "PageOrientation"
).generate_checker()

PageFormat = Literal[
    "", "a3", "A3", "a4", "A4", "a5", "A5", "letter", "Letter", "legal", "Legal"
]
check_page_format_type = LiteralCheckGenerator(
    PageFormat, "PageFormat"
).generate_checker()

MarginType = Literal["left", "right", "top", "bottom"]
check_margin_type_type = LiteralCheckGenerator(
    MarginType, "MarginType"
).generate_checker()

ClockMode = Literal["hms", "ms", "msreduced"]
check_clock_mode_type = LiteralCheckGenerator(ClockMode, "ClockMode").generate_checker()


def _get_name_of_check_type_function(musurgia_type: MusurgiaType) -> str:
    """
    >>> _get_name_of_check_type_function("MatrixIndex")
    'check_matrix_index_type'
    """
    return (
        "check"
        + "".join([f"_{x.lower()}" if x.isupper() else x for x in musurgia_type])
        + "_type"
    )


def get_check_musurgia_type(musurgia_type: MusurgiaType) -> Callable[[Any], bool]:
    if musurgia_type not in MUSURGIA_TYPES:
        raise TypeError(f"check_musurgia_type: invalid musurgia_type {musurgia_type}")
    check_function_name = _get_name_of_check_type_function(musurgia_type)
    try:
        func: Callable[[Any], bool] = globals()[check_function_name]
    except KeyError:
        raise AttributeError(
            f"get_check_musurgia_type: {check_function_name} does not exist"
        )
    return func


def check_type(
    v: Any,
    t: Union[type, str],
    function_name: Optional[str] = None,
    class_name: Optional[str] = None,
    method_name: Optional[str] = None,
    argument_name: Optional[str] = None,
    property_name: Optional[str] = None,
    class_attribute_name: Optional[str] = None,
) -> bool:
    """
    :param v: ``value`` to be checked.
    :param t: ``type``.
    :param function_name: see :obj:`MusurgiaTypeError`
    :param class_name: see :obj:`MusurgiaTypeError`
    :param method_name: see :obj:`MusurgiaTypeError`
    :param argument_name: see :obj:`MusurgiaTypeError`
    :param property_name: see :obj:`MusurgiaTypeError`
    :param class_attribute_name: see :obj:`MusurgiaTypeError`

    :raise: :obj:`MusurgiaTypeError`
    """

    def _create_error(message: Optional[str] = None) -> MusurgiaTypeError:
        if not message:
            return MusurgiaTypeError(
                v,
                t,
                function_name,
                class_name,
                method_name,
                argument_name,
                property_name,
                class_attribute_name,
            )
        else:
            return MusurgiaTypeError(
                None,
                None,
                function_name,
                class_name,
                method_name,
                argument_name,
                property_name,
                class_attribute_name,
                message,
            )

    if isinstance(t, type):
        if t is int and isinstance(v, bool):
            # in python bool is a subclass of int
            raise _create_error()
        elif not isinstance(v, t):
            raise _create_error()
    else:
        check_musurgia_type_type(t)
        try:
            get_check_musurgia_type(cast(MusurgiaType, t))(v)
        except TypeError as err:
            raise _create_error(message=str(err))
    return True



from pathlib import Path

from musicscore.score import Score
from musurgia.tests.utils_for_tests import XMLTestCase, create_test_fractal_musical_tree

path = Path(__file__)


class TestSimpleFractalMusicalTree(XMLTestCase):
    def setUp(self):
        self.ft = create_test_fractal_musical_tree()

    def test_midi_values(self):
        expected = """└── [72]
    ├── [72]
    │   ├── [72]
    │   ├── [72]
    │   │   ├── [72]
    │   │   ├── [72]
    │   │   ├── [72]
    │   │   └── [72]
    │   ├── [72]
    │   └── [72]
    │       ├── [72]
    │       ├── [72]
    │       ├── [72]
    │       └── [72]
    ├── [72]
    ├── [72]
    │   ├── [72]
    │   ├── [72]
    │   ├── [72]
    │   │   ├── [72]
    │   │   ├── [72]
    │   │   ├── [72]
    │   │   └── [72]
    │   └── [72]
    │       ├── [72]
    │       ├── [72]
    │       ├── [72]
    │       └── [72]
    └── [72]
        ├── [72]
        │   ├── [72]
        │   ├── [72]
        │   ├── [72]
        │   └── [72]
        ├── [72]
        │   ├── [72]
        │   ├── [72]
        │   ├── [72]
        │   └── [72]
        ├── [72]
        └── [72]
"""
        self.assertEqual(
            self.ft.get_tree_representation(
                key=lambda node: [midi.value for midi in node.get_chord_factory().midis]
            ),
            expected,
        )

    def test_create_chord(self):
        count = 0
        for node in self.ft.traverse():
            count += 1
            chord = node.get_chord_factory().create_chord()
            self.assertEqual(
                chord.quarter_duration, node.get_duration().get_quarter_duration()
            )

        self.assertEqual(count, 41)

    def test_simple_fractal_musical_tree_first_layer(self):
        score = Score()
        part = score.add_part("part-1")
        for node in self.ft.get_layer(1):
            part.add_chord(node.get_chord_factory().create_chord())

        with self.file_path(path, "simple_first_layer") as xml_path:
            score.export_xml(xml_path)

    def test_simple_fractal_musical_tree(self):
        score = self.ft.export_score()
        score.get_quantized = True
        with self.file_path(path, "simple") as xml_path:
            score.export_xml(xml_path)
